# pylint: disable=too-many-lines
"""Post processing of ansible-navigator configuration."""

from __future__ import annotations

import contextlib
import importlib
import logging
import os
import shlex
import shutil
import subprocess
import sys
import zoneinfo

from functools import partialmethod
from itertools import chain
from itertools import repeat
from pathlib import Path
from string import Formatter
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import ExitPrefix
from ansible_navigator.utils.definitions import LogMessage
from ansible_navigator.utils.functions import check_for_ansible
from ansible_navigator.utils.functions import check_playbook_type
from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import flatten_list
from ansible_navigator.utils.functions import oxfordcomma
from ansible_navigator.utils.functions import str2bool
from ansible_navigator.utils.functions import to_list

from .definitions import ApplicationConfiguration
from .definitions import CliParameters
from .definitions import Constants as C
from .definitions import Mode
from .definitions import ModeChangeRequest
from .definitions import PaeChangeRequest
from .definitions import SettingsEntry
from .definitions import VolumeMount
from .definitions import VolumeMountError


if TYPE_CHECKING:
    from collections.abc import Callable

strict_resolve = True
# Testing hack to allow errors trying to resolve /tmp/ansible.cfg during unit tests
if os.environ.get("PYTEST_VERSION", ""):
    strict_resolve = False


def _post_processor(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorate a post processing function.

    Args:
        func: The function to wrap

    Returns:
        The wrapped function
    """

    def wrapper(*args: Any, **kwargs: Any) -> PostProcessorReturn:
        name = kwargs["entry"].name
        before = str(kwargs["entry"].value.current)
        messages, exit_messages = func(*args, **kwargs)
        after = str(kwargs["entry"].value.current)
        changed = before != after
        messages.append(
            LogMessage(
                level=logging.DEBUG,
                message=f"Completed post processing for {name}. (changed={changed})",
            ),
        )
        if changed:
            messages.append(LogMessage(level=logging.DEBUG, message=f" before: '{before}'"))
            messages.append(LogMessage(level=logging.DEBUG, message=f" after: '{after}'"))
        return messages, exit_messages

    return wrapper


PostProcessorReturn = tuple[list[LogMessage], list[ExitMessage]]


class NavigatorPostProcessor:
    """Application post processor."""

    def __init__(self) -> None:
        """Initialize the post processor."""
        #: Volume mounts accumulated from post processing various config entries.
        #: These get processed towards the end, in the (delayed)
        #: :meth:`.execution_environment_volume_mounts` post-processor.
        self.extra_volume_mounts: list[VolumeMount] = []
        self._requested_mode: list[ModeChangeRequest] = []
        self._requested_pae: list[PaeChangeRequest] = []

    @staticmethod
    def _true_or_false(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process a boolean value.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        with contextlib.suppress(ValueError):
            entry.value.current = str2bool(entry.value.current)

        return messages, exit_messages

    @staticmethod
    @_post_processor
    def ansible_runner_artifact_dir(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process ansible_runner_artifact_dir path.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current is not C.NOT_SET:
            entry.value.current = str(expand_path(entry.value.current))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def ansible_runner_rotate_artifacts_count(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process ansible_runner_rotate_artifacts_count.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current is not C.NOT_SET:
            try:
                entry.value.current = int(entry.value.current)
            except ValueError as exc:
                exit_msg = f"Value should be valid integer. Failed with error {exc!s}"
                exit_messages.append(ExitMessage(message=exit_msg))
        return messages, exit_messages

    # Post process ansible_runner_write_job_events.
    ansible_runner_write_job_events = _true_or_false

    @staticmethod
    @_post_processor
    def ansible_runner_timeout(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process ansible_runner_timeout.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current is not C.NOT_SET:
            try:
                entry.value.current = int(entry.value.current)
            except ValueError as exc:
                exit_msg = f"Value should be valid integer. Failed with error {exc!s}"
                exit_messages.append(ExitMessage(message=exit_msg))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def collection_doc_cache_path(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process collection doc cache path.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        entry.value.current = str(expand_path(entry.value.current))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def cmdline(entry: SettingsEntry, config: ApplicationConfiguration) -> PostProcessorReturn:
        """Post process cmdline.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if isinstance(entry.value.current, str):
            entry.value.current = shlex.split(entry.value.current)
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def container_engine(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process container_engine.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current == "auto":
            choices = filter(lambda x: x != "auto", entry.choices)
            for choice in choices:
                if shutil.which(str(choice)):
                    entry.value.current = choice
                    entry.value.source = C.AUTO
                    break
        return messages, exit_messages

    @_post_processor
    def display_color(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process display_color.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.source == C.ENVIRONMENT_VARIABLE:
            entry.value.current = False
            message = f"{entry.environment_variable()} was set, set to {entry.value.current}"
            messages.append(LogMessage(level=logging.INFO, message=message))
            return messages, exit_messages
        return self._true_or_false(entry, config)

    # Post process editor_console.
    editor_console = _true_or_false

    @_post_processor
    def execution_environment(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        # pylint: disable=too-many-locals
        """Post process execution_environment.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages, exit_messages = self._true_or_false(entry, config)
        if entry.value.current is True:
            exit_msg = ""
            if config.container_engine == "auto":
                exit_msg = "No container engine could be found:"
            else:
                container_engine_location = shutil.which(config.container_engine)
                if container_engine_location is None:
                    exit_msg = "The specified container engine could not be found:"
            if exit_msg:
                exit_msg += f" '{config.container_engine}',"
                exit_msg += f" set by '{config.entry('container_engine').value.source.value}'"
                exit_messages.append(ExitMessage(message=exit_msg))
                ce_short: str | None = ""
                cli_parameters = config.entry("container_engine").cli_parameters
                if isinstance(cli_parameters, CliParameters):
                    ce_short = cli_parameters.short
                if isinstance(entry.cli_parameters, CliParameters):
                    entry_short = entry.cli_parameters.short
                else:
                    entry_short = None
                choices = config.entry("container_engine").choices
                if ce_short and entry_short:
                    hint = ""
                    if config.container_engine == "auto":
                        ce_choices = filter(lambda x: x != "auto", choices)
                        hint = f"Try installing {oxfordcomma(ce_choices, 'or')}"
                    elif config.container_engine in choices:
                        alternatives = filter(lambda x: x != config.container_engine, choices)
                        other = [f"{ce_short} {alt}" for alt in alternatives]
                        hint = f"Try installing '{config.container_engine}'"
                        hint += f", try again with {oxfordcomma(other, 'or')}"
                        hint += (
                            f" or even '{entry_short}"
                            " false' to disable the use of an execution environment"
                        )
                    if hint:
                        exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))

            # check for ``/dev/mqueue/`` when using podman because runner passes ipc=host
            # https://github.com/ansible/ansible-navigator/issues/610
            # except on Darwin (macOS)
            message_queue_path = "/dev/mqueue/"
            mqueue_is_not_dir = not Path(message_queue_path).is_dir()
            os_is_not_mac = sys.platform != "darwin"
            ce_is_podman = config.container_engine == "podman"
            if all((ce_is_podman, mqueue_is_not_dir, os_is_not_mac)):
                exit_msg = (
                    "Execution environment support while using podman requires a"
                    f" '{message_queue_path}' directory."
                )
                exit_messages.append(ExitMessage(message=exit_msg))
                hint = (
                    f"Try creating it with 'mkdir {message_queue_path}' or reference the"
                    " documentation for your operating system related to POSIX message queues."
                )
                exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))
        else:
            new_messages, new_exit_messages = check_for_ansible()
            messages.extend(new_messages)
            exit_messages.extend(new_exit_messages)

            if config.app == "exec" and isinstance(entry.cli_parameters, CliParameters):
                exit_msg = "The 'exec' subcommand requires execution environment support."
                exit_messages.append(ExitMessage(message=exit_msg))
                hint = (
                    f"Try again with '{entry.cli_parameters.short} true'"
                    " to enable the use of an execution environment."
                )
                exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))

        return messages, exit_messages

    @staticmethod
    @_post_processor
    def execution_environment_image(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process execution_environment_image.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if ":" not in entry.value.current:
            entry.value.current = f"{entry.value.current}:latest"
        return messages, exit_messages

    @_post_processor
    def execution_environment_volume_mounts(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process set_environment_variable.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        # pylint: disable=too-many-locals

        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        entry_name = entry.settings_file_path(prefix="")
        entry_source = entry.value.source
        volume_mounts: list[VolumeMount] = []

        if entry_source in (C.ENVIRONMENT_VARIABLE, C.USER_CLI):
            hint = (
                "Try again with format <source-path>:<destination-path>:<options>'."
                " Note: options is optional."
            )
            mount_strings = flatten_list(entry.value.current)
            for mount_str in mount_strings:
                src, dest, options, *left_overs = chain(mount_str.split(":"), repeat("", 3))
                if any(left_overs):
                    exit_msg = (
                        f"The following {entry_name} entry could not be parsed:"
                        f" {mount_str} ({entry_source.value})"
                    )
                    exit_messages.append(ExitMessage(message=exit_msg))
                    exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))

                try:
                    volume_mounts.append(
                        VolumeMount(
                            fs_source=src,
                            fs_destination=dest,
                            options_string=options,
                            settings_entry=entry_name,
                            source=entry_source,
                        ),
                    )
                except VolumeMountError as exc:
                    exit_msg = (
                        f"The following {entry_name} entry could not be parsed:"
                        f" {mount_str} ({entry.value.source.value}). Errors were found: {exc!s}"
                    )
                    exit_messages.append(ExitMessage(message=exit_msg))
                    exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))

        elif entry.value.source is C.USER_CFG:
            hint = (
                "The value of execution-environment.volume-mounts should be a list"
                " of dictionaries and valid keys are 'src', 'dest' and 'options'."
            )
            if not isinstance(entry.value.current, list):
                exit_msg = f"{entry_name} entries could not be parsed. ({entry_source.value})"
                exit_messages.append(ExitMessage(message=exit_msg))
                exit_msg = hint
                exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
            else:
                for volume_mount in entry.value.current:
                    try:
                        volume_mounts.append(
                            VolumeMount(
                                fs_source=volume_mount.get("src"),
                                fs_destination=volume_mount.get("dest"),
                                options_string=volume_mount.get("options", ""),
                                settings_entry=entry_name,
                                source=entry_source,
                            ),
                        )
                    except (AttributeError, VolumeMountError) as exc:
                        exit_msg = (
                            f"The following {entry_name} entry could not be parsed:  {volume_mount}"
                            f" ({entry_source.value}). Errors were found: {exc!s}"
                        )
                        exit_messages.append(ExitMessage(message=exit_msg))
                        exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))

        # Get out fast if we had any errors
        if exit_messages:
            return messages, exit_messages

        # New mounts were provided
        if volume_mounts:
            entry.value.current = [v.to_string() for v in volume_mounts]

        # Extra mounts were requested, these get added to either
        # new_mounts, C.PREVIOUS_CLI or C.NOT_SET
        if self.extra_volume_mounts:
            if not isinstance(entry.value.current, list):
                entry.value.current = []
            entry.value.current.extend(v.to_string() for v in self.extra_volume_mounts)

        # Finally, ensure the list has no duplicates
        if isinstance(entry.value.current, list):
            entry.value.current = sorted(set(entry.value.current), key=entry.value.current.index)

        return messages, exit_messages

    @staticmethod
    @_post_processor
    def container_options(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process container_options.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.source == C.ENVIRONMENT_VARIABLE:
            entry.value.current = to_list(entry.value.current)
        if entry.value.current is not C.NOT_SET:
            entry.value.current = flatten_list(entry.value.current)
        return messages, exit_messages

    @_post_processor
    def _disable_pae_and_enforce_stdout(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Disable playbook artifact creation and force mode stdout for a settings parameter.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages, exit_messages = self._true_or_false(entry, config)
        if entry.value.current is True:
            mode = Mode.STDOUT
            playbook_artifact_enable = False
            self._requested_mode.append(ModeChangeRequest(entry=entry.name, mode=mode))
            self._requested_pae.append(
                PaeChangeRequest(
                    entry=entry.name,
                    playbook_artifact_enable=playbook_artifact_enable,
                ),
            )
            message = (
                f"`{entry.name} requesting mode {mode.value} and pae as {playbook_artifact_enable}"
            )
            messages.append(LogMessage(level=logging.DEBUG, message=message))
        return messages, exit_messages

    # Post process for enable_prompts
    enable_prompts = _disable_pae_and_enforce_stdout

    # Post process for exec_shell
    exec_shell = _true_or_false

    @_post_processor
    def _forced_stdout(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
        subcommand: str,
    ) -> PostProcessorReturn:
        """Force mode stdout for a settings parameter.

        Args:
            entry: The current settings entry
            config: The full application configuration
            subcommand: The applicable subcommand

        Returns:
            An instance of the standard post process return object
        """
        messages, exit_messages = self._true_or_false(entry, config)

        if entry.value.current is True and config.app == subcommand:
            mode = Mode.STDOUT
            self._requested_mode.append(ModeChangeRequest(entry=entry.name, mode=mode))
            message = f"`{entry.name} requesting mode {mode.value}"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
        return messages, exit_messages

    help_builder = partialmethod(_forced_stdout, subcommand="builder")
    help_config = partialmethod(_forced_stdout, subcommand="config")
    help_doc = partialmethod(_forced_stdout, subcommand="doc")
    help_inventory = partialmethod(_forced_stdout, subcommand="inventory")

    # Post process for help_playbook
    help_playbook = _disable_pae_and_enforce_stdout

    @_post_processor
    def images_details(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process execution_environment_image_details.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        entry.value.current = flatten_list(entry.value.current)
        # In the case --default was provided, set it to everything
        if not entry.value.current:
            entry.value.current = entry.value.default

        # Only force mode stdout if from the CLI so the values can be kept in settings
        # or in an environment variable and used with only --mode stdout
        if entry.value.source is C.USER_CLI:
            mode = Mode.STDOUT
            self._requested_mode.append(ModeChangeRequest(entry=entry.name, mode=mode))
            message = f"`{entry.name} requesting mode {mode.value}"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
        return messages, exit_messages

    @_post_processor
    def _normalize_path(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Normalize path in config by resolving symlinks and relative paths.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        if entry.value.current is not C.NOT_SET:
            try:
                normalized = Path(entry.value.current).resolve(strict=strict_resolve).as_posix()
            except FileNotFoundError:  # pragma: no cover
                message = (
                    f"Path `{entry.value.current}` does not exist."
                    " Falling back to non-strict resolution."
                )
                messages.append(LogMessage(level=logging.WARNING, message=message))
                normalized = Path(entry.value.current).resolve().as_posix()
            if normalized != entry.value.current:
                entry.value.current = normalized
                message = f"`{entry.name} was normalized to {entry.value}."
                messages.append(LogMessage(level=logging.WARNING, message=message))
        return messages, exit_messages

    @_post_processor
    def config(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process config.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        return self._normalize_path(entry=entry, config=config)

    @staticmethod
    @_post_processor
    def inventory(entry: SettingsEntry, config: ApplicationConfiguration) -> PostProcessorReturn:
        """Post process inventory.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        if entry.value.current is C.NOT_SET:
            # Try the ansible.cfg file
            ansible_cfg = config.internals.ansible_configuration.contents
            if isinstance(ansible_cfg, dict):
                ansible_cfg_entry = ansible_cfg.get("defaults", {}).get("inventory")
                if isinstance(ansible_cfg_entry, str):
                    entry.value.current = [
                        str(expand_path(i)) for i in ansible_cfg_entry.split(",")
                    ]
                    entry.value.source = C.ANSIBLE_CFG
        else:
            # Use the specified
            flattened = flatten_list(entry.value.current)
            entry.value.current = []
            for inv_entry in flattened:
                if "," in inv_entry:
                    # host list
                    entry.value.current.append(inv_entry)
                else:
                    # file path
                    entry.value.current.append(str(expand_path(inv_entry)))

        # Something may be required
        app_match = config.app == "inventory"
        current_not_set = entry.value.current is C.NOT_SET
        not_help_inventory = config.entry("help_inventory").value.current is False

        if app_match and current_not_set and not_help_inventory:
            exit_msg = "An inventory is required when using the inventory subcommand"
            exit_messages.append(ExitMessage(message=exit_msg))
            if entry.cli_parameters:
                exit_msg = f"Try again with '{entry.cli_parameters.short} <path to inventory>'"
                exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))

        return messages, exit_messages

    @staticmethod
    @_post_processor
    def inventory_column(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process inventory_columns.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current is not C.NOT_SET:
            entry.value.current = flatten_list(entry.value.current)
        return messages, exit_messages

    def lintables(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post-process lintables.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        # If not using an EE, check for ansible-lint before we even pass off to
        # the lint action.
        if config.app == "lint" and not config.execution_environment:
            ansible_lint_location = shutil.which("ansible-lint")
            if ansible_lint_location is None:
                exit_messages.append(
                    ExitMessage(message="ansible-lint command could not be found."),
                )
                exit_messages.append(
                    ExitMessage(
                        message=(
                            "Try 'pip install ansible-lint' or consider using an execution "
                            "environment which provides ansible-lint."
                        ),
                        prefix=ExitPrefix.HINT,
                    ),
                )
            else:
                try:
                    subprocess.run(
                        "ansible-lint --version",  # noqa:S607
                        shell=True,
                        check=True,
                    )
                except subprocess.CalledProcessError:
                    exit_messages.append(
                        ExitMessage(
                            message=("ansible-lint does not seem to be installed correctly."),
                        ),
                    )
                    exit_messages.append(
                        ExitMessage(
                            message=(
                                "Ensure the command `ansible-lint --version` can be run prior to"
                                " using ansible-navigator"
                            ),
                            prefix=ExitPrefix.HINT,
                        ),
                    )

        if isinstance(entry.value.current, str) and config.app == "lint":
            entry_name = entry.settings_file_path(prefix="")
            entry_source = entry.value.source
            source = str(expand_path(entry.value.current))
            try:
                mount = VolumeMount(
                    fs_source=source,
                    fs_destination=source,
                    settings_entry=entry_name,
                    source=entry_source,
                    options_string="",
                )
            except VolumeMountError as ex:
                exit_messages.append(
                    ExitMessage(
                        message=f"Error mounting lintable into execution environment: {ex}",
                    ),
                )
            else:
                self.extra_volume_mounts.append(mount)

        return messages, exit_messages

    # Post process log_append.
    log_append = _true_or_false

    @staticmethod
    @_post_processor
    def log_file(entry: SettingsEntry, config: ApplicationConfiguration) -> PostProcessorReturn:
        """Post process log_file.

        If the parent directory for the log file cannot be created and is writable.
        If not restore to default, this will allow the writing of log messages
        even if application initialization results in a sys.exit condition

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        entry.value.current = str(expand_path(entry.value.current))
        try:
            Path(Path(entry.value.current).parent).mkdir(parents=True, exist_ok=True)
            Path(entry.value.current).touch()
        except OSError as exc:
            exit_msgs = [
                (
                    f"Failed to create log file {entry.value.current}"
                    f" specified in '{entry.value.source.value}'"
                ),
            ]
            exit_msgs.append(f"The error was: {exc!s}")
            exit_messages.extend(ExitMessage(message=exit_msg) for exit_msg in exit_msgs)
            entry.value.current = entry.value.default
            entry.value.source = C.DEFAULT_CFG
            exit_msg = f"Log file set to default location: {entry.value.current}."
            exit_messages.append(ExitMessage(message=exit_msg))
            if entry.cli_parameters:
                exit_msg = (
                    f"Try again with '{entry.cli_parameters.short}"
                    " ~/ansible-navigator.log' to place it in your home directory"
                )
                exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
        return messages, exit_messages

    @staticmethod
    def _load_subcommand_action(
        subcommand_name: str,
        config: ApplicationConfiguration,
        messages: list[LogMessage],
    ) -> Any:
        """Load the action module for the given subcommand name.

        Args:
            subcommand_name: The name of the subcommand to load
            config: The full application configuration
            messages: The list of log messages to append to

        Returns:
            The loaded subcommand action, or None if not found
        """
        for action_package_name in config.internals.action_packages:
            try:
                action_package = importlib.import_module(action_package_name)
            except ImportError as exc:
                message = f"Unable to load action package: '{action_package_name}': {exc!s}"
                messages.append(LogMessage(level=logging.ERROR, message=message))
                continue
            try:
                if hasattr(action_package, "get"):
                    valid_pkg_name = subcommand_name.replace("-", "_")
                    return action_package.get(valid_pkg_name)
                break
            except (AttributeError, ModuleNotFoundError) as exc:
                message = f"Unable to load subcommand '{subcommand_name}' from"
                message += f" action package: '{action_package_name}': {exc!s}"
                messages.append(LogMessage(level=logging.DEBUG, message=message))
        return None

    def _apply_mode_change_requests(
        self,
        entry: SettingsEntry,
        messages: list[LogMessage],
    ) -> None:
        """Apply any accumulated mode change requests to the entry.

        Args:
            entry: The current settings entry
            messages: The list of log messages to append to

        Raises:
            ValueError: When more than 1 distinct mode change is requested
        """
        mode_set = {request.mode for request in self._requested_mode}
        if len(mode_set) == 1:
            requested = self._requested_mode[0]
            auto_mode = requested.mode.value
            if auto_mode != entry.value.current:
                entry_mode_change_msgs = (
                    f"Parameter '{requested.entry}' required mode '{auto_mode!s}'.",
                    f"Mode will be set to '{auto_mode}'",
                )
                messages.extend(
                    LogMessage(level=logging.INFO, message=msg) for msg in entry_mode_change_msgs
                )
                entry.value.current = auto_mode
                entry.value.source = C.AUTO
        elif len(mode_set) > 1:
            msg = f"Conflicting mode requests: {self._requested_mode}"
            raise ValueError(msg)

    @_post_processor
    def mode(self, entry: SettingsEntry, config: ApplicationConfiguration) -> PostProcessorReturn:
        """Post process mode.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        subcommand_name = config.subcommand(config.app).name

        # Post initialization of mode processing is not needed since switching modes after
        # is not supported
        if not config.internals.initializing:
            return messages, exit_messages

        subcommand_action = self._load_subcommand_action(subcommand_name, config, messages)

        if subcommand_action is None:
            exit_msg = f"Unable to find an action for '{subcommand_name}', tried: "
            exit_msg += oxfordcomma(config.internals.action_packages, "and")
            exit_messages.append(ExitMessage(message=exit_msg))
            return messages, exit_messages

        if entry.value.current == "interactive" and os.environ.get("TERM") is None:
            exit_msg = "The TERM environment variable must be set for mode 'interactive'"
            exit_messages.append(ExitMessage(message=exit_msg))
            exit_msg = (
                "Try again after setting the TERM environment variable"
                " (e.g., 'export TERM=xterm256color') from the command line"
            )
            exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))

        # Check if the mode interactive is available for the subcommand
        # mode stdout is always available since the action base class has a `run_stdout`
        subcommand_stdout_only = not hasattr(subcommand_action, "run")
        if entry.value.current == "interactive" and subcommand_stdout_only:
            subcommand_mode_change_msgs = (
                f"Subcommand '{config.app}' does not support mode 'interactive'.",
                "Mode set to 'interactive'",
            )
            messages.extend(
                LogMessage(level=logging.INFO, message=msg) for msg in subcommand_mode_change_msgs
            )
            entry.value.current = "stdout"
            entry.value.source = C.AUTO

        # Check if any other entry has requested a mode change different than current
        self._apply_mode_change_requests(entry, messages)
        return messages, exit_messages

    # Post process osc4.
    osc4 = _true_or_false

    @_post_processor
    def plugin_name(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process plugin_name.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        is_not_subcommand_match = config.entry("app").value.current != "doc"
        if is_not_subcommand_match:
            # Subcommand is different
            return messages, exit_messages

        is_help = config.entry("help_doc").value.current is True
        if is_help:
            # Help is requested
            return messages, exit_messages

        # Determine if mode stdout should be set
        if config.initial is None:
            force_stdout_for = {
                "-l",  # List plugins
                "--list",  # List plugins
                "-F",  # List plugin files
                "--list_files",  # List plugin files
                "-s",  # Print a snippet
                "--snippet",  # Print a snippet
                "--metadata-dump",  # Print a metadata dump
            }
            cmdline = config.entry("cmdline").value.current
            if isinstance(cmdline, list) and force_stdout_for.intersection(cmdline):
                mode = Mode.STDOUT
                self._requested_mode.append(ModeChangeRequest(entry=entry.name, mode=mode))
                message = f"`{entry.name} requesting mode {mode.value}"
                messages.append(LogMessage(level=logging.DEBUG, message=message))
                return messages, exit_messages

        is_not_set = entry.value.current is C.NOT_SET
        if is_not_set:
            # A plugin name is required
            exit_msg = "A plugin name or other parameter is required when using the doc subcommand"
            exit_messages.append(ExitMessage(message=exit_msg))
            exit_msg = "Try again with 'doc <plugin_name>'"
            exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def pass_environment_variable(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process pass_environment_variable.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.current is not C.NOT_SET:
            entry.value.current = flatten_list(entry.value.current)
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def playbook(entry: SettingsEntry, config: ApplicationConfiguration) -> PostProcessorReturn:
        """Post process playbook.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if (
            config.app == "run"
            and entry.value.current is C.NOT_SET
            and config.entry("help_playbook").value.current is False
        ):
            exit_msg = "A playbook is required when using the run subcommand"
            exit_messages.append(ExitMessage(message=exit_msg))
            exit_msg = "Try again with 'run <playbook name>'"
            exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
            return messages, exit_messages
        if check_playbook_type(entry.value.current) == "file" and isinstance(
            entry.value.current,
            str,
        ):
            entry.value.current = str(expand_path(entry.value.current))
        return messages, exit_messages

    @_post_processor
    def playbook_artifact_enable(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process playbook_artifact_enable.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Raises:
            ValueError: When more than 1 pae changes requests are
                present, shouldn't happen

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        # Check if any other entry has requested a pae change different than current
        pae_set = {request.playbook_artifact_enable for request in self._requested_pae}
        if len(pae_set) == 1:
            requested = self._requested_pae[0]
            auto_pae = requested.playbook_artifact_enable
            if auto_pae != entry.value.current:
                pae_change_msgs = (
                    f"Parameter '{requested.entry}' required pae '{auto_pae!s}'.",
                    f"Pae will be set to '{auto_pae}'",
                )
                messages.extend(
                    LogMessage(level=logging.INFO, message=msg) for msg in pae_change_msgs
                )
                entry.value.current = auto_pae
        elif len(pae_set) > 1:
            msg = f"Conflicting pae requests: {self._requested_pae}"
            raise ValueError(msg)
        with contextlib.suppress(ValueError):
            entry.value.current = str2bool(entry.value.current)

        return messages, exit_messages

    @staticmethod
    @_post_processor
    def playbook_artifact_replay(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process playbook_artifact_replay.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if config.app == "replay" and entry.value.current is C.NOT_SET:
            exit_msg = "An playbook artifact file is required when using the replay subcommand"
            exit_messages.append(ExitMessage(message=exit_msg))
            exit_msg = "Try again with 'replay <path to playbook artifact>'"
            exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
            return messages, exit_messages

        if isinstance(entry.value.current, str):
            entry.value.current = str(expand_path(entry.value.current))

        if config.app == "replay" and not Path(entry.value.current).is_file():
            exit_msg = f"The specified playbook artifact could not be found: {entry.value.current}"
            exit_messages.append(ExitMessage(message=exit_msg))
            exit_msg = "Try again with 'replay <valid path to playbook artifact>'"
            exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
            return messages, exit_messages
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def playbook_artifact_save_as(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process playbook_artifact_save_as.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        # literal_text, fname, format_spec, conversion
        found = {f for _, f, _, _ in Formatter().parse(entry.value.current) if f}
        available = {f for _, f, _, _ in Formatter().parse(entry.value.default) if f}
        non_defaults_also_available = {"playbook_status"}

        available.update(non_defaults_also_available)
        unknown = found - available
        if not unknown:
            return messages, exit_messages
        exit_msg = (
            f"The playbook artifact file name '{entry.value.current}', set by"
            f" {entry.value.source.value.lower()}, has unrecognized variables:"
            f" {oxfordcomma(unknown, 'and')}"
        )
        exit_messages.append(ExitMessage(message=exit_msg))
        exit_msg = f"Try again with only {oxfordcomma(available, 'and/or')}"
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def pull_arguments(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process ``pull_arguments``.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.source == C.ENVIRONMENT_VARIABLE:
            entry.value.current = to_list(entry.value.current)
        if entry.value.current is not C.NOT_SET:
            entry.value.current = flatten_list(entry.value.current)
        return messages, exit_messages

    settings_effective = partialmethod(_forced_stdout, subcommand="settings")
    settings_sample = partialmethod(_forced_stdout, subcommand="settings")
    settings_sources = partialmethod(_forced_stdout, subcommand="settings")

    @_post_processor
    def settings_schema(
        self,
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Force mode stdout for schema parameter.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        if entry.value.source is not C.DEFAULT_CFG and config.app == "settings":
            mode = Mode.STDOUT
            self._requested_mode.append(ModeChangeRequest(entry=entry.name, mode=mode))
            message = f"`{entry.name} requesting mode {mode.value}"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def set_environment_variable(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process set_environment_variable.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []
        if entry.value.source in [
            C.ENVIRONMENT_VARIABLE,
            C.USER_CLI,
        ]:
            flattened = flatten_list(entry.value.current)
            set_envs = {}
            for env_var_pair in flattened:
                parts = env_var_pair.split("=")
                if len(parts) == 2:
                    set_envs[parts[0]] = parts[1]
                else:
                    exit_msg = (
                        "The following set-environment-variable"
                        f" entry could not be parsed: {env_var_pair}"
                    )
                    exit_messages.append(ExitMessage(message=exit_msg))
                    if entry.cli_parameters:
                        exit_msg = f"Try again with '{entry.cli_parameters.short} MY_VAR=my_value'"
                        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
            entry.value.current = set_envs
        if entry.value.source is not C.NOT_SET:
            entry.value.current = {k: str(v) for k, v in entry.value.current.items()}
        return messages, exit_messages

    @staticmethod
    @_post_processor
    def time_zone(
        entry: SettingsEntry,
        config: ApplicationConfiguration,
    ) -> PostProcessorReturn:
        """Post process ``time_zone``.

        Args:
            entry: The current settings entry
            config: The full application configuration

        Returns:
            An instance of the standard post process return object
        """
        messages: list[LogMessage] = []
        exit_messages: list[ExitMessage] = []

        exit_msg = (
            f"The specified time zone '{entry.value.current}', set by"
            f" {entry.value.source.value.lower()}, "
        )
        if isinstance(entry.value.current, str):
            if entry.value.current in ("UTC", "local"):
                return messages, exit_messages
            available_timezones = sorted(zoneinfo.available_timezones())
            if entry.value.current in available_timezones:
                return messages, exit_messages
            exit_msg += "could not be found."
        else:
            exit_msg += (
                f"must be a string but was found to be a '{type(entry.value.current).__name__}'."
            )
        exit_messages.append(ExitMessage(message=exit_msg))
        exit_msg = "Please try again with a valid IANA time zone."
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
        return messages, exit_messages



from __future__ import annotations

from re import Match  # noqa: F401
from typing import TypeVar


T = TypeVar("T")


def uniquely_constructed(t: T) -> T:
    """Avoid tuple.__hash__ for "singleton" constructed objects."""
    t.__hash__ = object.__hash__  # type: ignore[assignment]
    return t



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple

from ._types import Protocol
from .fchainmap import FChainMap
from .reg import ERR_REG
from .reg import _Reg
from .reg import _RegSet
from .reg import do_regset
from .reg import expand_escaped
from .reg import make_reg
from .region import Region
from .region import Regions
from .state import State
from .tokenize import tokenize
from .utils import uniquely_constructed


if TYPE_CHECKING:
    from re import Match

    from .compiler import Compiler
    from .region import Scope

Captures = tuple[tuple[int, "_Rule"], ...]  # declared later


def _split_name(s: str | None) -> tuple[str, ...]:
    if s is None:
        return ()
    return tuple(s.split())


class CompiledRule(Protocol):
    @property
    def name(self) -> tuple[str, ...]: ...

    def start(
        self,
        compiler: Compiler,
        match: Match[str],
        state: State,
    ) -> tuple[State, bool, Regions]: ...

    def search(
        self,
        compiler: Compiler,
        state: State,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[State, int, bool, Regions] | None: ...


class Entry(NamedTuple):
    scope: tuple[str, ...]
    rule: CompiledRule
    start: tuple[str, int]
    reg: _Reg = ERR_REG
    boundary: bool = False


class _Rule(Protocol):
    """hack for recursive types python/mypy#731."""

    @property
    def name(self) -> tuple[str, ...]: ...

    @property
    def match(self) -> str | None: ...

    @property
    def begin(self) -> str | None: ...

    @property
    def end(self) -> str | None: ...

    @property
    def while_(self) -> str | None: ...

    @property
    def content_name(self) -> tuple[str, ...]: ...

    @property
    def captures(self) -> Captures: ...

    @property
    def begin_captures(self) -> Captures: ...

    @property
    def end_captures(self) -> Captures: ...

    @property
    def while_captures(self) -> Captures: ...

    @property
    def include(self) -> str | None: ...

    @property
    def patterns(self) -> tuple[_Rule, ...]: ...

    @property
    def repository(self) -> FChainMap[str, _Rule]: ...


class CompiledRegsetRule(CompiledRule, Protocol):
    @property
    def regset(self) -> _RegSet: ...

    @property
    def u_rules(self) -> tuple[_Rule, ...]: ...


@uniquely_constructed
class EndRule(NamedTuple):
    name: tuple[str, ...]
    content_name: tuple[str, ...]
    begin_captures: Captures
    end_captures: Captures
    end: str
    regset: _RegSet
    u_rules: tuple[_Rule, ...]

    def start(
        self,
        compiler: Compiler,
        match: Match[str],
        state: State,
    ) -> tuple[State, bool, Regions]:
        scope = state.cur.scope + self.name
        next_scope = scope + self.content_name

        boundary = match.end() == len(match.string)
        reg = make_reg(expand_escaped(match, self.end))
        start = (match.string, match.start())
        state = state.push(Entry(next_scope, self, start, reg, boundary))
        regions = _captures(compiler, scope, match, self.begin_captures)
        return state, True, regions

    def _end_ret(
        self,
        compiler: Compiler,
        state: State,
        pos: int,
        m: Match[str],
    ) -> tuple[State, int, bool, Regions]:
        ret = []
        if m.start() > pos:
            ret.append(Region(pos, m.start(), state.cur.scope))
        ret.extend(_captures(compiler, state.cur.scope, m, self.end_captures))
        # this is probably a bug in the grammar, but it pushed and popped at
        # the same position.
        # we'll advance the highlighter by one position to get past the loop
        # this appears to be what vs code does as well
        if state.entries[-1].start == (m.string, m.end()):
            ret.append(Region(m.end(), m.end() + 1, state.cur.scope))
            end = m.end() + 1
        else:
            end = m.end()
        return state.pop(), end, False, tuple(ret)

    def search(
        self,
        compiler: Compiler,
        state: State,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[State, int, bool, Regions] | None:
        end_match = state.cur.reg.search(line, pos, first_line, boundary)
        if end_match is not None and end_match.start() == pos:
            return self._end_ret(compiler, state, pos, end_match)
        if end_match is None:
            idx, match = self.regset.search(line, pos, first_line, boundary)
            return do_regset(idx, match, self, compiler, state, pos)
        idx, match = self.regset.search(line, pos, first_line, boundary)
        if match is None or end_match.start() <= match.start():
            return self._end_ret(compiler, state, pos, end_match)
        return do_regset(idx, match, self, compiler, state, pos)


@uniquely_constructed
class MatchRule(NamedTuple):
    name: tuple[str, ...]
    captures: Captures

    def start(
        self,
        compiler: Compiler,
        match: Match[str],
        state: State,
    ) -> tuple[State, bool, Regions]:
        scope = state.cur.scope + self.name
        return state, False, _captures(compiler, scope, match, self.captures)

    def search(
        self,
        _compiler: Compiler,
        _state: State,
        _line: str,
        _pos: int,
        _first_line: bool,
        _boundary: bool,
    ) -> tuple[State, int, bool, Regions] | None:
        msg = f"unreachable {self}"
        raise AssertionError(msg)


@uniquely_constructed
class PatternRule(NamedTuple):
    name: tuple[str, ...]
    regset: _RegSet
    u_rules: tuple[_Rule, ...]

    def start(
        self,
        _compiler: Compiler,
        _match: Match[str],
        _state: State,
    ) -> tuple[State, bool, Regions]:
        msg = f"unreachable {self}"
        raise AssertionError(msg)

    def search(
        self,
        compiler: Compiler,
        state: State,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[State, int, bool, Regions] | None:
        idx, match = self.regset.search(line, pos, first_line, boundary)
        return do_regset(idx, match, self, compiler, state, pos)


@uniquely_constructed
class Rule(NamedTuple):
    name: tuple[str, ...]
    match: str | None
    begin: str | None
    end: str | None
    while_: str | None
    content_name: tuple[str, ...]
    captures: Captures
    begin_captures: Captures
    end_captures: Captures
    while_captures: Captures
    include: str | None
    patterns: tuple[_Rule, ...]
    repository: FChainMap[str, _Rule]

    @classmethod
    def _parse_captures_key(
        cls,
        dct: dict[str, Any],
        key: str,
        repository: FChainMap[str, _Rule],
    ) -> Captures:
        """Parse a captures key from the grammar dict.

        Args:
            dct: The grammar dictionary.
            key: The captures key name.
            repository: The rule repository.

        Returns:
            The parsed captures tuple.
        """
        if key in dct:
            return tuple(
                (int(k), Rule.make(v, repository))
                for k, v in sorted(dct[key].items(), key=lambda x: int(x[0]))
            )
        return ()

    @classmethod
    def make(cls, dct: dict[str, Any], parent_repository: FChainMap[str, _Rule]) -> _Rule:
        if "repository" in dct:
            repository_dct: dict[str, _Rule] = {}
            repository = FChainMap(parent_repository, repository_dct)
            for k, sub_dct in dct["repository"].items():
                repository_dct[k] = Rule.make(sub_dct, repository)
        else:
            repository = parent_repository

        name = _split_name(dct.get("name"))
        match = dct.get("match")
        begin = dct.get("begin")
        end = dct.get("end")
        while_ = dct.get("while")
        content_name = _split_name(dct.get("contentName"))

        captures = cls._parse_captures_key(dct, "captures", repository)
        begin_captures = cls._parse_captures_key(dct, "beginCaptures", repository)
        end_captures = cls._parse_captures_key(dct, "endCaptures", repository)
        while_captures = cls._parse_captures_key(dct, "whileCaptures", repository)

        if begin is not None and end is None and while_ is None:
            end = "$impossible^"

        if begin and end and captures:
            begin_captures = end_captures = captures
            captures = ()
        elif begin and while_ and captures:
            begin_captures = while_captures = captures
            captures = ()

        include = dct.get("include")

        if "patterns" in dct:
            patterns = tuple(Rule.make(d, repository) for d in dct["patterns"])
        else:
            patterns = ()

        return cls(
            name=name,
            match=match,
            begin=begin,
            end=end,
            while_=while_,
            content_name=content_name,
            captures=captures,
            begin_captures=begin_captures,
            end_captures=end_captures,
            while_captures=while_captures,
            include=include,
            patterns=patterns,
            repository=repository,
        )


@uniquely_constructed
class WhileRule(NamedTuple):
    name: tuple[str, ...]
    content_name: tuple[str, ...]
    begin_captures: Captures
    while_captures: Captures
    while_: str
    regset: _RegSet
    u_rules: tuple[_Rule, ...]

    def start(
        self,
        compiler: Compiler,
        match: Match[str],
        state: State,
    ) -> tuple[State, bool, Regions]:
        scope = state.cur.scope + self.name
        next_scope = scope + self.content_name

        boundary = match.end() == len(match.string)
        reg = make_reg(expand_escaped(match, self.while_))
        start = (match.string, match.start())
        entry = Entry(next_scope, self, start, reg, boundary)
        state = state.push_while(self, entry)
        regions = _captures(compiler, scope, match, self.begin_captures)
        return state, True, regions

    def continues(
        self,
        compiler: Compiler,
        state: State,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[int, bool, Regions] | None:
        match = state.cur.reg.match(line, pos, first_line, boundary)
        if match is None:
            return None

        ret = _captures(compiler, state.cur.scope, match, self.while_captures)
        return match.end(), True, ret

    def search(
        self,
        compiler: Compiler,
        state: State,
        line: str,
        pos: int,
        first_line: bool,
        boundary: bool,
    ) -> tuple[State, int, bool, Regions] | None:
        idx, match = self.regset.search(line, pos, first_line, boundary)
        return do_regset(idx, match, self, compiler, state, pos)


def _backtrack_capture(
    compiler: Compiler,
    ret: list[Region],
    start: int,
    end: int,
    group_s: str,
    rule: CompiledRule,
) -> None:
    """Handle a capture group that starts before the current position.

    Args:
        compiler: The grammar compiler.
        ret: The regions list to splice into.
        start: The start of the capture group.
        end: The end of the capture group.
        group_s: The captured string.
        rule: The compiled rule for the capture.
    """
    j = len(ret) - 1
    while j > 0 and start < ret[j - 1].end:
        j -= 1

    oldtok = ret[j]
    newtok = []
    if start > oldtok.start:
        newtok.append(oldtok._replace(end=start))

    newtok.extend(_inner_capture_parse(compiler, start, group_s, oldtok.scope, rule))

    if end < oldtok.end:
        newtok.append(oldtok._replace(start=end))
    ret[j : j + 1] = newtok


def _captures(
    compiler: Compiler,
    scope: Scope,
    match: Match[str],
    captures: Captures,
) -> Regions:
    ret: list[Region] = []
    pos, pos_end = match.span()
    for i, u_rule in captures:
        try:
            group_s = match[i]
        except IndexError:
            continue
        if not group_s:
            continue

        rule = compiler.compile_rule(u_rule)
        start, end = match.span(i)
        if start < pos:
            _backtrack_capture(compiler, ret, start, end, match[i], rule)
        else:
            if start > pos:
                ret.append(Region(pos, start, scope))
            ret.extend(_inner_capture_parse(compiler, start, match[i], scope, rule))
            pos = end

    if pos < pos_end:
        ret.append(Region(pos, pos_end, scope))
    return tuple(ret)


def _inner_capture_parse(
    compiler: Compiler,
    start: int,
    s: str,
    scope: Scope,
    rule: CompiledRule,
) -> Regions:
    state = State.root(Entry(scope + rule.name, rule, (s, 0)))
    _, regions = tokenize(compiler, state, s, first_line=False)
    return tuple(r._replace(start=r.start + start, end=r.end + start) for r in regions)



"""Screen refresh implementation."""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """Screen refresh implementation."""

    KEGEX = r"^KEY_F\(5\)$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the refresh action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the screen refresh request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        # Just in case the user switched tasks with +,- etc
        # change previous, since this interaction is on the stack
        if interaction.content:
            app.steps.previous.index = interaction.action.value



"""Functionality related to the retrieval of packaged data files."""

from enum import Enum
from enum import auto
from importlib import resources
from pathlib import Path


class ImageEntry(Enum):
    """A mapping of ID to sequence within the image dockerfile.

    The values are not currently used.
    """

    DEFAULT_EE = auto()
    SMALL_IMAGE = auto()
    PULLABLE_IMAGE = auto()

    def get(self, app_name: str) -> str:
        """Retrieve an image name from the packaged container file.

        Args:
            app_name: The name of the application.

        Returns:
            The default execution environment image.
        """
        file_contents = retrieve_content(app_name=app_name, filename="images_dockerfile")
        from_line = (
            line
            for line in file_contents.splitlines()
            if line.startswith("FROM") and self.name in line
        )

        image = next(from_line).split()[1]
        return image


def retrieve_content(filename: str, app_name: str = "ansible_navigator") -> str:
    """Retrieve the content of a packaged data file.

    Args:
        app_name: The name of the application.
        filename: The name of the file to retrieve.

    Returns:
        The content of the file.
    """
    data_directory = "data"
    package = f"{app_name}.{data_directory}"

    with resources.files(package).joinpath(filename).open("r", encoding="utf-8") as fh:
        content = fh.read()

    return content


def path_to_file(filename: str, app_name: str = "ansible_navigator") -> Path:
    """Retrieve the path to a packaged data file.

    Args:
        app_name: The name of the application.
        filename: The name of the file to retrieve.

    Returns:
        The path to the file.
    """
    data_directory = "data"
    package = f"{app_name}.{data_directory}"

    path = Path(str(resources.files(package).joinpath(filename)))

    return path



# cspell:ignore fqcn,FQCN

"""This action provides ansible-lint results through ansible-navigator.

Internally, it works by using ansible-runner to execute ansible-lint (optionally
in an execution environment). When running ansible-lint, it passes
``-f codeclimate`` which requests JSON output on stdout. The JSON output contains
a list of issues which we then report in the ansible-navigator UI.

We allow users to dig into an individual issue using the standard :<lineno>
commands, to learn more about the issue.

The full specification for ansible-lint's use of JSON (for the codeclimate
formatter) can be found in src/ansiblelint/formatters/__init__.py in the
ansible-lint codebase.
"""

from __future__ import annotations

import json
import shlex

from collections.abc import Mapping
from datetime import datetime
from datetime import timezone
from enum import IntEnum
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.runner.command import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import Color
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Decoration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import error_notification
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import success_notification
from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import remove_ansi
from ansible_navigator.utils.functions import time_stamp_for_file

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


class Severity(IntEnum):
    """A mapping from ansible-lint severity to an integer represented internally.

    Primarily used for sorting.
    """

    INFO = 10
    MINOR = 20
    MAJOR = 30
    CRITICAL = 40
    BLOCKER = 50

    # If ansible-lint ever gives us something we don't expect, return this.
    # In practice, we reverse-sort, so this will always be last.
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> Severity:
        """Return unknown if ansible-lint ever returns something unexpected.

        Args:
            value: The value

        Returns:
            A severity unknown
        """
        return Severity.UNKNOWN


def severity_to_color(severity: str) -> int:
    """Convert severity to curses colors.

    Args:
        severity: The severity to convert to a color

    Returns:
        A color for the severity
    """
    if severity == "minor":
        return Color.MAGENTA
    if severity == "major":
        return Color.YELLOW
    if severity in ("critical", "blocker"):
        return Color.RED
    if severity == "info":
        return Color.CYAN
    return Color.BLACK


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Color the menu.

    Args:
        colno: The column number
        colname: The column name
        entry: The current content entry

    Returns:
        The foreground and background color
    """
    return (severity_to_color(entry["severity"]), Color.BLACK)


def content_heading(obj: dict[Any, Any], screen_w: int) -> CursesLines:
    """Generate the content heading.

    Args:
        obj: The content for which the heading will be generated
        screen_w: The current screen width

    Returns:
        The content heading
    """
    check_line = f"Message: {obj['check_name']}"
    location = f"Location: {obj['issue_path']}"
    fill_characters = screen_w - len(location) + 1
    location_line = f"{location}{' ' * fill_characters}"

    line_1_part_1 = CursesLinePart(
        column=0,
        string=check_line,
        color=severity_to_color(obj["severity"]),
        decoration=Decoration.NORMAL,
    )
    line_2_part_1 = CursesLinePart(
        column=0,
        string=location_line,
        color=severity_to_color(obj["severity"]),
        decoration=Decoration.UNDERLINE,
    )
    return CursesLines((CursesLine((line_1_part_1,)), CursesLine((line_2_part_1,))))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out internal keys.

    Args:
        obj: The content from which the content keys will be filtered

    Returns:
        The content without the internal keys
    """
    ignored_keys = ("fingerprint",)
    return {k: v for k, v in obj.items() if not k.startswith("__") and k not in ignored_keys}


def massage_issue(issue: dict[Any, Any]) -> dict[Any, Any]:
    """Massage an issue by injecting some useful keys with strings for rendering.

    Args:
        issue: The issue reported

    Returns:
        The issue reformatted
    """
    massaged = issue.copy()
    massaged["__severity"] = massaged["severity"].capitalize()
    # Version 6.1 and before of ansible-lint used this syntax
    # "check_name": "[fqcn-builtins] Use FQCN for builtin actions."
    # Version 6.2 and later use this syntax
    # check_name": "fqcn-builtins", "description": "Use FQCN for builtin actions."
    if issue["check_name"].startswith("["):
        massaged["__message"] = issue["check_name"].split("]")[1].strip()
    else:
        massaged["__message"] = issue["description"]
    massaged["__path"] = expand_path(issue["location"]["path"])
    if "lines" in issue["location"] and isinstance(issue["location"]["lines"]["begin"], Mapping):
        massaged["__line"] = issue["location"]["lines"]["begin"]["line"]
    elif "positions" in issue["location"] and isinstance(
        issue["location"]["positions"]["begin"],
        Mapping,
    ):
        massaged["__line"] = issue["location"]["positions"]["begin"]
    else:
        massaged["__line"] = issue["location"]["lines"]["begin"]
    massaged["issue_path"] = f"{massaged['__path']}:{massaged['__line']}"
    return massaged


MENU_COLUMNS = [
    "__severity",
    "__message",
    "__path",
    "__line",
]


@actions.register
class Action(ActionBase):
    """Run the lint subcommand."""

    KEGEX = r"^lint(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the action.

        Args:
            args: The current application configuration
        """
        self._modification_times_last_updated: datetime
        self._success_notification_shown: bool = False
        self._issues_menu = Step(
            name="issues_menu",
            step_type="menu",
            columns=MENU_COLUMNS,
            value=[],
            select_func=self._build_issue_content,
        )
        self._current_issue_count = 0
        super().__init__(args=args, logger_name=__name__, name="lint")

    @property
    def is_interactive(self) -> bool:
        """Determine if interactive.

        Returns:
            An indication is running in interactive mode
        """
        return bool(self._args.mode == "interactive")

    def _fatal(self, msg: str) -> None:
        """Show a notification if a fatal error has occurred.

        Args:
            msg: The message to display

        Raises:
            RuntimeError: A runtime error if not mode interactive
        """
        self._logger.error(msg)

        if self.is_interactive:
            notification = error_notification(messages=[msg])
            self._interaction.ui.show_form(notification)
        else:
            raise RuntimeError(msg)

    def _run_runner(self) -> tuple[str, str, int]:
        """Spin up runner to run ansible-lint, either in an exec env or not.

        Returns:
            The output, errors and return code
        """
        kwargs = {
            "container_engine": self._args.container_engine,
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": self._args.set_environment_variable,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
            "host_cwd": str(Path.cwd()),
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs["container_volume_mounts"] = self._args.execution_environment_volume_mounts

        if isinstance(self._args.container_options, list):
            kwargs["container_options"] = self._args.container_options

        cmd_args = ["-qq", "--offline"]

        if self.is_interactive:
            cmd_args += [
                "--nocolor",
                "-f",
                "codeclimate",
            ]
        elif not self._args.display_color:
            cmd_args.append("--nocolor")

        if isinstance(self._args.lint_config, str):
            cmd_args += [
                "-c",
                self._args.lint_config,
            ]

        if isinstance(self._args.lintables, str):
            # Does it actually exist?
            if not Path(self._args.lintables).exists():
                self._fatal(f"The given path `{self._args.lintables}` does not exist.")
                return "", "", 1
            cmd_args.append(self._args.lintables)

        kwargs["cmdline"] = cmd_args

        runner = Command(executable_cmd="ansible-lint", **kwargs)
        return runner.run()

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``doc`` request for mode stdout.

        Returns:
            A message and return code
        """
        self._logger.debug("lint requested in stdout mode")
        _output, _error, return_code = self._run_runner()
        return RunStdoutReturn(message="", return_code=return_code)

    @staticmethod
    def _pull_out_json_or_fatal(stdout: str) -> str | None:
        """Attempt to pull out JSON line from ansible-lint raw output.

        Note that stdout and stderr get munged together by docker/podman, so we
        need to do some trickery to try to figure out the actual JSON line vs,
        say, ansible warnings.

        Args:
            stdout: The stdout from the lint invocation

        Returns:
            The json string
        """
        # We want the last (non empty) line of output. This should hopefully be
        # the JSON we need.
        for line in reversed(stdout.splitlines()):
            if not line:
                continue
            return line
        return None

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``lint`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("lint requested")
        self._prepare_to_run(app, interaction)

        updated = self._update_args(
            ["lint"] + shlex.split(interaction.action.match.groupdict()["params"] or ""),
        )

        if not updated:
            self._prepare_to_exit(interaction)
            return None

        notification = nonblocking_notification(messages=["Linting, this may take a minute..."])
        interaction.ui.show_form(notification)

        self._build_issues_menu()
        if self._current_issue_count == 0:
            self._prepare_to_exit(interaction)
            return None

        self.steps.append(self._issues_menu)

        while True:
            self.update()

            if self._current_issue_count == 0:
                break

            self._interaction.ui.update_status(
                status=f"Issues: {self._current_issue_count}",
                status_color=self._max_severity_color,
            )

            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def update(self) -> None:
        """Request calling app update, and update modification time if needed."""
        self._calling_app.update()

        # Do this only every 2 seconds
        if (datetime.now(timezone.utc) - self._modification_times_last_updated).total_seconds() > 2:
            rerun_lint = self._rerun_needed()
            if rerun_lint:
                self._build_issues_menu()

    @property
    def _max_severity_color(self) -> int:
        """Determine the color of the maximum severity issue.

        Returns:
            The color
        """
        max_severity = max(Severity[i["severity"].upper()].value for i in self._issues_menu.value)
        return severity_to_color(Severity(max_severity).name.lower())

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_issues_menu(self) -> None:
        """Build the menu of all issues.

        Returns:
            Indication of success
        """
        output, _error, return_code = self._run_runner()
        self._logger.debug("Output from ansible-lint run (rc=%d): %s", return_code, output)

        # ansible-lint failed
        if return_code != 0 and "ansible-lint: No such file or directory" in output:
            installed_or_ee = (
                "in the execution environment you are using"
                if self._args.execution_environment
                else "installed"
            )
            self._fatal(
                "ansible-lint executable could not be found. Ensure 'ansible-lint' "
                f"is {installed_or_ee} and try again.",
            )
            self._current_issue_count = 0
            return

        # Could not extract json from response
        out_without_warnings = self._pull_out_json_or_fatal(output)
        if out_without_warnings is None:
            self._current_issue_count = 0
            return

        # De-serialization of json failed
        try:
            raw_issues = json.loads(out_without_warnings)
        except json.JSONDecodeError as exc:
            self._logger.debug("Failed to parse 'ansible-lint' JSON response: %s", str(exc))
            messages = ["Could not parse 'ansible-lint' output."]
            without_ansi = remove_ansi(output)
            messages.extend(without_ansi.splitlines())
            notification = error_notification(messages)
            self._interaction.ui.show_form(notification)
            self._current_issue_count = 0
            return

        # No issues were found
        if len(raw_issues) == 0:
            notification = success_notification(messages=["Congratulations, no lint issues found!"])
            self._interaction.ui.show_form(notification)

        # Update menu
        issues = [massage_issue(issue) for issue in raw_issues]
        self._current_issue_count = len(issues)
        self._issues_menu.value = sorted(
            issues,
            key=lambda i: Severity[i["severity"].upper()],
            reverse=True,
        )

        # Update new content timestamps
        _rerun_lint = self._rerun_needed()

    def _build_issue_content(self) -> Step:
        """Build the content for one plugin.

        Returns:
            The plugin's content
        """
        return Step(
            name="issue_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _rerun_needed(self) -> bool:
        """Add a modification timestamp to each issue, check for changes.

        Returns:
            Indication if lint should be rerun
        """
        rerun_lint = False
        path_timestamps: dict[str, tuple[float | None, str | None]] = {}
        for issue in self._issues_menu.value:
            path = issue["__path"]
            if path not in path_timestamps:
                path_timestamps[path] = time_stamp_for_file(path, self._args.time_zone)

            unix_ts, iso_ts = path_timestamps[path]
            previous_ts = issue.get("__last_modified")

            issue["last_modified"] = iso_ts
            issue["__last_modified"] = unix_ts

            if unix_ts is not None and (previous_ts is None or unix_ts > previous_ts):
                rerun_lint = True

        self._modification_times_last_updated = datetime.now(timezone.utc)
        if rerun_lint:
            self._logger.debug("Files modified")
        return rerun_lint



"""Conditional imports related to python versions."""

import importlib.metadata as importlib_metadata
import sys


# https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
from typing import TypeAlias


if sys.version_info >= (3, 11):
    from importlib.resources.abc import Traversable
else:
    from importlib.abc import Traversable


__all__ = ["Traversable", "TypeAlias", "importlib_metadata"]



"""Herein lies the ability to have ansible-runner run a command in a synchronous manner."""

from __future__ import annotations

from ansible_runner import run_command

from .command_base import CommandBase


class Command(CommandBase):
    """A runner wrapper."""

    def run(self) -> tuple[str, str, int]:
        """Run command.

        Returns:
            Output, error, and error code
        """
        self.generate_run_command_args()
        out, err, ret_code = run_command(**self._runner_args)
        return out, err, ret_code



"""``:open`` command implementation."""

from __future__ import annotations

import curses
import logging
import os

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.content_defs import ContentBase
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.content_defs import ContentType
from ansible_navigator.content_defs import ContentView
from ansible_navigator.content_defs import SerializationFormat
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.serialize import serialize_write_temp_file

from . import _actions as actions


if TYPE_CHECKING:
    from collections.abc import Callable
    from re import Pattern
    from types import TracebackType

    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
    from ansible_navigator.ui_framework import Interaction
    from ansible_navigator.ui_framework import Menu


class SuspendCurses:
    """Context Manager to temporarily leave curses mode."""

    def __enter__(self) -> None:
        """Close the curses window."""
        curses.endwin()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Open the curses window.

        Args:
            exc_type: The exception class
            exc_val: The type of exception
            traceback: Report of all information related to the
                exception
        """
        new_scr = curses.initscr()
        new_scr.refresh()
        curses.doupdate()


@actions.register
class Action:
    """``:open`` command implementation."""

    KEGEX = r"^o(?:pen)?(\s(?P<requested>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:open`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def _transform_menu(
        self,
        menu: Menu,
        menu_filter: Callable[..., Pattern[str] | None],
        serialization_format: SerializationFormat | None,
    ) -> list[dict[Any, Any] | ContentBase[Any]]:
        """Convert a menu into structured data.

        Args:
            menu: The current menu showing
            menu_filter: The effective menu filter
            serialization_format: The current serialization format

        Returns:
            The menu converted to a structured data
        """
        self._logger.debug("menu is showing, open that")
        menu_entries: list[dict[Any, Any] | ContentBase[Any]] = []
        for entry in menu.current:
            if isinstance(entry, ContentBase) and serialization_format is not None:
                entry = entry.asdict(
                    content_view=ContentView.FULL,
                    serialization_format=serialization_format,
                )
            menu_entries.append(entry)

        menu_entries = [
            {remove_dbl_un(k): v for k, v in c.items() if k in menu.columns} for c in menu_entries
        ]
        menu_filter_result = menu_filter()
        if menu_filter_result:
            menu_entries = [
                e
                for e in menu_entries
                if menu_filter_result.search(" ".join(str(v) for v in e.values()))  # type: ignore[union-attr]
            ]
        return menu_entries

    @staticmethod
    def _assess_requested_is_file(requested: str) -> tuple[Path | None, int | None]:
        """Determine is the user requested string is a file.

        Args:
            requested: The string requested at the ``:`` prompt

        Returns:
            None, None or file name and line number
        """
        parts = requested.rsplit(":", 1)
        possible_filename = Path(parts[0])
        if possible_filename.is_file():
            file_name = possible_filename
            try:
                line_number = int(parts[1:][0]) if parts[1:] else 0
            except ValueError:
                line_number = 0
            return file_name, line_number
        return None, None

    def _open_a_file(
        self,
        file_name: Path,
        line_number: int,
        editor_console: bool,
        editor_command: str,
    ) -> None:
        """Open a file using the editor.

        Args:
            file_name: The name of the file to open
            line_number: The line number the file should be opened at
            editor_console: Indicates if the editor is console based
            editor_command: The templatable editor command
        """
        command = editor_command.format(filename=file_name, line_number=line_number)

        self._logger.debug("Command: %s", command)
        if isinstance(command, str):
            if editor_console:
                with SuspendCurses():
                    os.system(command)  # noqa:S605
            else:
                os.system(command)  # noqa:S605

    @staticmethod
    def _persist_content(content: ContentType, content_format: ContentFormat) -> Path:
        """Write content to a temporary file.

        Args:
            content: the content to write
            content_format: The format of the content

        Returns:
            The path to the file
        """
        file_name = serialize_write_temp_file(
            content=content,
            content_view=ContentView.NORMAL,
            content_format=content_format,
        )
        return file_name

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:open`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            Nothing
        """
        self._logger.debug("open requested")

        editor_command = self._args.editor_command
        editor_console = self._args.editor_console
        requested = interaction.action.match.groupdict()["requested"]
        content_format = interaction.ui.content_format()
        serialization_format = content_format.value.serialization

        if requested:
            requested_file_name, requested_line_number = self._assess_requested_is_file(requested)
            if isinstance(requested_file_name, Path) and isinstance(requested_line_number, int):
                self._open_a_file(
                    file_name=requested_file_name,
                    line_number=requested_line_number,
                    editor_console=editor_console,
                    editor_command=editor_command,
                )
                return

            temp_file_name = self._persist_content(content=requested, content_format=content_format)
            temp_line_number = 0
            self._open_a_file(
                file_name=temp_file_name,
                line_number=temp_line_number,
                editor_console=editor_console,
                editor_command=editor_command,
            )
            return

        if interaction.content:
            content = interaction.content.showing
        elif interaction.menu:
            content = self._transform_menu(
                menu=interaction.menu,
                menu_filter=interaction.ui.menu_filter,
                serialization_format=serialization_format,
            )
        else:
            return

        content_file_name = self._persist_content(content=content, content_format=content_format)
        content_line_number = 0
        self._open_a_file(
            file_name=content_file_name,
            line_number=content_line_number,
            editor_console=editor_console,
            editor_command=editor_command,
        )



"""``:write`` command implementation."""

import logging
import re

from pathlib import Path
from typing import Any

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentView
from ansible_navigator.content_defs import SerializationFormat
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.serialize import serialize_write_file

from . import _actions as actions


JSON_EXTENSION = ".json"


@actions.register
class Action:
    """``:write`` command implementation."""

    KEGEX = r"^w(?:rite)?(?P<force>!)?\s+(?P<append>>>)?\s*(?P<filename>.+)$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:write`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def _determine_file_mode(
        self,
        filename: str,
        match: dict[str, str],
    ) -> str | None:
        """Determine the file open mode based on append/force flags.

        Args:
            filename: The resolved filename path
            match: The regex match groupdict from the interaction

        Returns:
            The file mode string ('a' or 'w'), or None if the operation should be aborted
        """
        if match["append"]:
            if not Path(filename).exists() and not match["force"]:
                self._logger.warning(
                    "Append operation failed because %s does not exist, force with !",
                    filename,
                )
                return None
            return "a"

        if Path(filename).exists() and not match["force"]:
            self._logger.warning(
                "Write operation failed because %s exists, force with !",
                filename,
            )
            return None
        return "w"

    @staticmethod
    def _resolve_content(interaction: Interaction) -> Any:
        """Resolve the content to write from the interaction.

        Args:
            interaction: The interaction from the user

        Returns:
            The content object to be written
        """
        if interaction.content:
            return interaction.content.showing

        if interaction.menu:
            obj = [
                {remove_dbl_un(k): v for k, v in c.items() if k in interaction.menu.columns}
                for c in interaction.menu.current
            ]
            if interaction.ui.menu_filter():
                obj = [
                    e
                    for e in obj
                    if interaction.ui.menu_filter().search(" ".join(str(v) for v in e.values()))
                ]
            return obj

        return None

    @staticmethod
    def _write_content(
        obj: Any,
        filename: str,
        file_mode: str,
        write_as: str,
    ) -> None:
        """Write the resolved content to a file in the determined format.

        Args:
            obj: The content to write
            filename: The target file path
            file_mode: The file open mode ('a' or 'w')
            write_as: The file extension determining serialization format
        """
        if write_as == ".txt":
            file = expand_path(filename)
            with file.open(file_mode, encoding="utf-8") as fh:
                fh.write(obj)
        elif write_as == ".yaml":
            serialize_write_file(
                content=obj,
                content_view=ContentView.NORMAL,
                file=Path(filename),
                file_mode=file_mode,
                serialization_format=SerializationFormat.YAML,
            )
        elif write_as == JSON_EXTENSION:
            serialize_write_file(
                content=obj,
                content_view=ContentView.NORMAL,
                file=Path(filename),
                file_mode=file_mode,
                serialization_format=SerializationFormat.JSON,
            )

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:write`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            Nothing
        """
        self._logger.debug("write requested as %s", interaction.action.value)
        match = interaction.action.match.groupdict()
        filename = str(expand_path(match["filename"]))

        file_mode = self._determine_file_mode(filename, match)
        if file_mode is None:
            return

        obj = self._resolve_content(interaction)
        if obj is None:
            self._logger.error("No menu or content found to write")
            return

        if isinstance(obj, str):
            write_as = ".txt"
        elif re.match(r"^.*\.y(?:a)?ml$", filename):
            write_as = ".yaml"
        elif filename.endswith(".json"):
            write_as = ".json"
        else:
            write_as = interaction.ui.content_format().value.file_extension

        self._write_content(obj, filename, file_mode, write_as)
        self._logger.info("Wrote to '%s' with mode '%s' as '%s'", filename, file_mode, write_as)



"""Constants for ansible-navigator."""

import importlib.resources as importlib_resources

from ansible_navigator.utils.compatibility import Traversable


PKG_NAME: str = "ansible_navigator"
DATA_DIR: Traversable = importlib_resources.files(PKG_NAME).joinpath("data")

THEME_DIR: Traversable = DATA_DIR.joinpath("themes")
THEME_NAME: str = "dark_vs.json"
THEME_PATH: Traversable = THEME_DIR.joinpath(THEME_NAME)

TERMINAL_COLORS: str = "terminal_colors.json"
TERMINAL_COLORS_PATH: Traversable = THEME_DIR.joinpath(TERMINAL_COLORS)

GRAMMAR_DIR: Traversable = DATA_DIR.joinpath("grammar")



"""Print pretty colors."""

from __future__ import annotations

import logging
import math
import os

from sys import stdout
from typing import TYPE_CHECKING

from ansible_navigator.constants import GRAMMAR_DIR
from ansible_navigator.constants import THEME_PATH
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.content_defs import ContentType
from ansible_navigator.content_defs import ContentView
from ansible_navigator.ui_framework.colorize import Colorize
from ansible_navigator.ui_framework.colorize import rgb_to_ansi
from ansible_navigator.utils.serialize import serialize


if TYPE_CHECKING:
    from ansible_navigator.ui_framework.curses_defs import SimpleLinePart


logger = logging.getLogger(__name__)


def color_bits() -> int:
    """Determine to number of color bit the terminal can support.

    Returns:
        The number of color bits
    """
    color_term = os.environ.get("COLORTERM", "").strip().lower()
    if color_term in ("truecolor", "24bit"):
        return 24
    term = os.environ.get("TERM", "").strip().lower()
    _term_name, _hyphen, colors = term.rpartition("-")
    colors = colors.replace("color", "")
    try:
        return int(math.log2(int(colors)))
    except (ValueError, TypeError):
        return 4


def _ansi_code_for_part(
    color: tuple[int, int, int] | None,
    term_color_bits: int,
) -> str:
    """Compute the ANSI escape code for a line part.

    Args:
        color: The RGB color tuple, or None for default.
        term_color_bits: The number of color bits the terminal supports.

    Returns:
        The ANSI escape code string.
    """
    if term_color_bits == 24:
        if color is None:
            return "\033[38;2;255;255;255m"
        red, green, blue = color
        return f"\033[38;2;{red};{green};{blue}m"
    if color is None:
        ansi_color = 1
    else:
        red, green, blue = color
        number_of_colors = 2**term_color_bits
        ansi_color = rgb_to_ansi(red, green, blue, number_of_colors)
    return f"\033[38;5;{ansi_color}m"


def color_lines(term_color_bits: int, tokenized: list[list[SimpleLinePart]]) -> str:
    """Transform tokenized lines to ANSI lines.

    Args:
        term_color_bits: The number of color bits the terminal supports
        tokenized: The tokenized content

    Returns:
        The ANSI string
    """
    lines = []
    for line in tokenized:
        printable = ""
        for line_part in line:
            ansi_code = _ansi_code_for_part(line_part.color, term_color_bits)
            printable += f"{ansi_code}{line_part.chars}\033[m"
        lines.append(printable)
    return "".join(lines)


def tokenize(
    content_format: ContentFormat,
    serialized: str,
) -> list[list[SimpleLinePart]]:
    """Serialize and tokenize an object.

    Args:
        content_format: The format type
        serialized: The serialized content

    Returns:
        A list of list of line parts
    """
    colorizer = Colorize(
        grammar_dir=GRAMMAR_DIR,
        theme_path=THEME_PATH,
    )

    tokenized = colorizer.render(doc=serialized, scope=content_format.value.scope)
    return tokenized


def print_to_stdout(
    content: ContentType,
    content_format: ContentFormat,
    use_color: bool,
) -> None:
    """Print some colored output to stdout.

    Args:
        content: The content to print out.
        content_format: The content_format
        use_color: Indicates if color should be used
    """
    serialization_format = content_format.value.serialization
    if serialization_format:
        serialized = str(
            serialize(
                content=content,
                content_view=ContentView.NORMAL,
                serialization_format=serialization_format,
            ),
        )
        output = serialized
    else:
        output = str(content)

    if use_color and not stdout.isatty():
        logger.debug("Color requested, but device is not a TTY")
        use_color = False

    if use_color:
        tokenized = tokenize(
            content_format=content_format,
            serialized=output,
        )
        terminal_color_bits = color_bits()
        output = color_lines(terminal_color_bits, tokenized)

    print(output)



# cspell:ignore ftype, chksum
"""Catalog collections within the execution environment."""

from __future__ import annotations

import argparse
import ast
import hashlib
import json
import multiprocessing
import os
import re
import subprocess
import sys

from collections import Counter
from collections import OrderedDict
from datetime import datetime
from datetime import timezone
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import yaml

from ansible import __version__ as ansible_version
from ansible import plugins
from ansible.utils.plugin_docs import get_docstring
from yaml.error import YAMLError


if TYPE_CHECKING:
    from collections.abc import Iterator


try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader  # type: ignore

# Import from the source tree whenever possible. When running
# in an execution environment, and therefore not type checking
# import from the key_value_store which was injected in the path.
# The TYPE_CHECKING conditional prevents mypy from attempting the
# import and causing an import error.
try:
    from ansible_navigator.utils.key_value_store import KeyValueStore
except ImportError:
    if not TYPE_CHECKING:
        from key_value_store import KeyValueStore


PROCESSES = (multiprocessing.cpu_count() - 1) or 1


class CollectionCatalog:
    """A collection cataloger."""

    def __init__(self, directories: list[Path]) -> None:
        """Initialize the collection cataloger.

        Args:
            directories: A list of directories that may contain
                collections
        """
        self._directories: list[Path] = directories
        self._collections: OrderedDict[str, dict[Any, Any]] = OrderedDict()
        self._errors: list[dict[str, str]] = []
        self._messages: list[str] = []

    def _catalog_plugins(self, collection: dict[Any, Any]) -> None:
        """Catalog the plugins within a collection.

        Args:
            collection: Details describing the collection
        """
        file_checksums = {}

        file_manifest_file = collection.get("file_manifest_file", {}).get("name")
        if file_manifest_file:
            file_path = Path(collection["path"], file_manifest_file)
            if file_path.exists():
                with file_path.open(encoding="utf-8") as fh:
                    try:
                        loaded = json.load(fh)
                        file_checksums = {v["name"]: v for v in loaded["files"]}
                    except (JSONDecodeError, KeyError) as exc:
                        self._errors.append({"path": str(file_path), "error": str(exc)})

        exempt = ["action", "module_utils", "doc_fragments"]
        plugin_directory = Path(collection["path"], "plugins")
        if plugin_directory.is_dir():
            plugin_dirs = [
                plugin_dir
                for plugin_dir in plugin_directory.iterdir()
                if plugin_dir.is_dir() and plugin_dir.name not in exempt
            ]
            # builtin modules are found in a sibling of the plugin directory
            if collection["known_as"] == "ansible.builtin":
                plugin_dirs.append(Path(collection["path"], "modules"))

            for plugin_dir in plugin_dirs:
                plugin_type = plugin_dir.name
                if plugin_type == "modules":
                    plugin_type = "module"
                files = list(plugin_dir.glob("**/*.py")) + list(plugin_dir.glob("**/*.yml"))
                filenames = iter(files)
                self._process_plugin_dir(
                    plugin_type,
                    filenames,
                    file_checksums,
                    collection,
                )

    def _load_role_argspecs(
        self,
        role: dict[str, str | dict[str, Any]],
        role_directory: Path,
        collection_name: str,
    ) -> None:
        """Load argument specs for a role. Non-fatal on error.

        Args:
            role: The role dict to populate.
            role_directory: Path to the role directory.
            collection_name: The collection name.
        """
        argspec_name = "argument_specs.yml"
        argspec_path = role_directory / "meta" / argspec_name
        role["argument_specs"] = {}
        role["argument_specs_path"] = ""
        error = {"path": str(argspec_path)}
        try:
            with argspec_path.open(encoding="utf-8") as fh:
                parsed = yaml.load(fh, Loader=SafeLoader)
                if not isinstance(parsed, dict) or "argument_specs" not in parsed:
                    error["error"] = f"Malformed {argspec_name} for role in {collection_name}."
                    self._errors.append(error)
                else:
                    role["argument_specs"] = parsed["argument_specs"]
                    role["argument_specs_path"] = str(argspec_path)
        except KeyError:
            error["error"] = f"Malformed {argspec_name} for role in {collection_name}."
            self._errors.append(error)
        except FileNotFoundError:
            error["error"] = f"Failed to find {argspec_name} for role in {collection_name}."
            self._errors.append(error)
        except YAMLError:
            error["error"] = f"Failed to load {argspec_name} for role in {collection_name}."
            self._errors.append(error)

    def _load_role_defaults(
        self,
        role: dict[str, str | dict[str, Any]],
        role_directory: Path,
        collection_name: str,
    ) -> bool:
        """Load defaults for a role.

        Args:
            role: The role dict to populate.
            role_directory: Path to the role directory.
            collection_name: The collection name.

        Returns:
            True if a fatal error occurred.
        """
        defaults_name = "main.yml"
        defaults_path = role_directory / "defaults" / defaults_name
        role["defaults"] = {}
        role["defaults_path"] = ""
        error = {"path": str(defaults_path)}
        try:
            with defaults_path.open(encoding="utf-8") as fh:
                parsed = yaml.load(fh, Loader=SafeLoader)
                role["defaults"] = parsed if isinstance(parsed, dict) else {}
                role["defaults_path"] = str(defaults_path)
        except FileNotFoundError:
            pass
        except YAMLError:
            error["error"] = f"Failed to load {defaults_name} for role in {collection_name}."
            self._errors.append(error)
            return True
        return False

    def _load_role_info(
        self,
        role: dict[str, str | dict[str, Any]],
        role_directory: Path,
        collection_name: str,
    ) -> bool:
        """Load meta/main.yml info for a role.

        Args:
            role: The role dict to populate.
            role_directory: Path to the role directory.
            collection_name: The collection name.

        Returns:
            True if a fatal error occurred.
        """
        meta_name = "main.yml"
        meta_path = role_directory / "meta" / meta_name
        role["info"] = {}
        role["info_path"] = ""
        error = {"path": str(meta_path)}
        try:
            with meta_path.open(encoding="utf-8") as fh:
                parsed = yaml.load(fh, Loader=SafeLoader)
                if not isinstance(parsed, dict):
                    error["error"] = f"Malformed {meta_name} for role in {collection_name}."
                    self._errors.append(error)
                    return True
                role["info"] = parsed
                role["info_path"] = str(meta_path)
        except FileNotFoundError:
            error["error"] = f"Failed to find {meta_name} for role in {collection_name}."
            self._errors.append(error)
            return True
        except YAMLError:
            error["error"] = f"Failed to load {meta_name} for role in {collection_name}."
            self._errors.append(error)
            return True
        return False

    def _load_role_readme(
        self,
        role: dict[str, str | dict[str, Any]],
        role_directory: Path,
        collection_name: str,
    ) -> bool:
        """Load README.md for a role.

        Args:
            role: The role dict to populate.
            role_directory: Path to the role directory.
            collection_name: The collection name.

        Returns:
            True if a fatal error occurred.
        """
        readme_name = "README.md"
        readme_path = role_directory / readme_name
        role["readme"] = ""
        role["readme_path"] = ""
        error = {"path": str(readme_path)}
        try:
            with readme_path.open(encoding="utf-8") as fh:
                role["readme"] = fh.read()
                role["readme_path"] = str(readme_path)
        except FileNotFoundError:
            error["error"] = f"Failed to find {readme_name} for role in {collection_name}."
            self._errors.append(error)
            return True
        return False

    def _catalog_roles(self, collection: dict[str, Any]) -> None:
        """Catalog the roles within a collection.

        Args:
            collection: Details describing the collection
        """
        collection_name: str = collection["known_as"]
        collection["roles"] = []
        roles_directory = Path(collection["path"], "roles")
        if not roles_directory.is_dir():
            return

        for role_directory in roles_directory.iterdir():
            if not role_directory.is_dir():
                continue
            role: dict[str, str | dict[str, Any]] = {
                "short_name": role_directory.name,
                "full_name": f"{collection_name}.{role_directory.name}",
            }
            self._load_role_argspecs(role, role_directory, collection_name)
            error_cataloging_role = self._load_role_defaults(role, role_directory, collection_name)
            error_cataloging_role = (
                self._load_role_info(role, role_directory, collection_name) or error_cataloging_role
            )
            error_cataloging_role = (
                self._load_role_readme(role, role_directory, collection_name)
                or error_cataloging_role
            )
            if not error_cataloging_role:
                collection["roles"].append(role)

    @staticmethod
    def _generate_checksum(file_path: Path, relative_path: Path) -> dict[str, Any]:
        """Generate a standard checksum for a file.

        Args:
            file_path: The path to the file to generate a checksum for
            relative_path: The relative path within the collection
                directory structure

        Returns:
            Details about the file, including the checksum
        """
        sha256_hash = hashlib.sha256()
        with file_path.open("rb") as fh:
            for byte_block in iter(lambda: fh.read(4096), b""):
                sha256_hash.update(byte_block)
        res = {
            "name": relative_path,
            "ftype": "file",
            "chksum_type": "sha256",
            "chksum_sha256": sha256_hash.hexdigest(),
            "format": 1,
        }
        return res

    def _process_plugin_dir(
        self,
        plugin_type: str,
        filenames: Iterator[Path],
        file_checksums: dict[str, dict[Any, Any]],
        collection: dict[Any, Any],
    ) -> None:
        """Process each plugin within one plugin directory.

        Args:
            plugin_type: The type of plugins
            filenames: The filenames of the plugins
            file_checksums: The checksums for the plugin files
            collection: The details of the collection
        """
        for filename in filenames:
            if str(filename.name).startswith("__"):
                continue

            relative_path = Path(filename).relative_to(collection["path"])
            checksum_dict = file_checksums.get(str(relative_path))
            if not checksum_dict:
                checksum_dict = self._generate_checksum(filename, relative_path)
            checksum = checksum_dict[f"chksum_{checksum_dict['chksum_type']}"]
            collection["plugin_checksums"][checksum] = {
                "path": str(relative_path),
                "type": plugin_type,
            }

    def _load_collection_meta(self, directory_path: Path) -> dict[str, Any] | None:
        """Load collection metadata from MANIFEST.json or galaxy.yml.

        Args:
            directory_path: The path to a single collection directory.

        Returns:
            The collection dict, or None if no metadata could be loaded.
        """
        manifest_file = directory_path / "MANIFEST.json"
        galaxy_file = directory_path / "galaxy.yml"
        if manifest_file.exists():
            with manifest_file.open(encoding="utf-8") as fh:
                try:
                    collection = json.load(fh)
                except JSONDecodeError:
                    self._errors.append(
                        {"path": str(manifest_file), "error": "failed to load MANIFEST.json"},
                    )
                else:
                    if not isinstance(collection, dict):
                        self._errors.append(
                            {"path": str(manifest_file), "error": "failed to load MANIFEST.json"},
                        )
                    else:
                        collection["meta_source"] = "MANIFEST.json"
                        return collection
        elif galaxy_file.exists():
            with galaxy_file.open(encoding="utf-8") as fh:
                try:
                    parsed = yaml.load(fh, Loader=SafeLoader)
                except YAMLError:
                    self._errors.append(
                        {"path": str(galaxy_file), "error": "failed to load galaxy.yml"},
                    )
                else:
                    if not isinstance(parsed, dict):
                        self._errors.append(
                            {"path": str(galaxy_file), "error": "failed to load galaxy.yml"},
                        )
                    else:
                        collection = {"collection_info": parsed}
                        collection["meta_source"] = "galaxy.yml"
                        return collection
        return None

    def _finalize_collection(
        self,
        collection: dict[str, Any],
        directory_path: Path,
    ) -> None:
        """Set derived fields and load runtime.yml for a collection.

        Args:
            collection: The collection dict to finalize.
            directory_path: The path to the collection directory.
        """
        collection_name = f"{collection['collection_info']['namespace']}"
        collection_name += f".{collection['collection_info']['name']}"
        collection["known_as"] = collection_name
        collection["plugin_checksums"] = {}
        collection["path"] = str(directory_path)

        runtime_file = directory_path / "meta" / "runtime.yml"
        collection["runtime"] = {}
        if runtime_file.exists():
            with runtime_file.open(encoding="utf-8") as fh:
                try:
                    parsed_runtime = yaml.load(fh, Loader=SafeLoader)
                    collection["runtime"] = (
                        parsed_runtime if isinstance(parsed_runtime, dict) else {}
                    )
                except YAMLError as exc:
                    self._errors.append({"path": str(runtime_file), "error": str(exc)})

    def _one_path(self, directory: Path) -> None:
        """Process the contents of an <...>/ansible_collections/ directory.

        Args:
            directory: The path to collections directory to walk and
                load
        """
        for directory_path in directory.glob("*/*/"):
            collection = self._load_collection_meta(directory_path)
            if collection:
                self._finalize_collection(collection, directory_path)
                self._collections[collection["path"]] = collection
            else:
                msg = (
                    f"collection path '{directory_path}' is ignored as it does not"
                    " have 'MANIFEST.json' and/or 'galaxy.yml' file(s)."
                )
                self._messages.append(msg)

    def _find_shadows(self) -> None:
        """Determine which collections are hidden by another installation of the same."""
        collection_list = list(self._collections.values())
        counts = Counter([collection["known_as"] for collection in collection_list])
        for idx, (collection_path, o_collection) in reversed(
            list(enumerate(self._collections.items())),
        ):
            self._collections[collection_path]["hidden_by"] = []
            if counts[o_collection["known_as"]] > 1:
                for i_collection in reversed(collection_list[0:idx]):
                    if i_collection["known_as"] == o_collection["known_as"]:
                        self._collections[collection_path]["hidden_by"].insert(
                            0,
                            i_collection["path"],
                        )

    def process_directories(self) -> tuple[dict[Any, Any], list[dict[str, str]]]:
        """Process each parent directory.

        Returns:
            All collections found and any errors
        """
        for directory in self._directories:
            collection_directory = directory / "ansible_collections"
            if collection_directory.exists():
                self._one_path(collection_directory)
        self.add_pseudo_builtin()
        for collection in self._collections.values():
            self._catalog_plugins(collection)
            self._catalog_roles(collection)
        self._find_shadows()
        return self._collections, self._errors

    def add_pseudo_builtin(self) -> None:
        """Add the pseudo builtin collection."""
        collection: dict[str, str | list[str] | dict[Any, Any]] = {}
        collection["known_as"] = "ansible.builtin"
        collection["plugin_checksums"] = {}
        collection["path"] = str(Path(plugins.__file__).parents[1])
        collection["runtime"] = {}
        collection["meta_source"] = "None"
        collection["collection_info"] = {"version": ansible_version}
        self._collections[str(collection["path"])] = collection
        msg = f"Added ansible.builtin from: {collection['path']}"
        self._messages.append(msg)


def worker(
    pending_queue: multiprocessing.Queue[Any],
    completed_queue: multiprocessing.Queue[Any],
) -> None:
    """Extract the documentation from a plugin, place in completed queue.

    Args:
        pending_queue: A queue with plugins to process
        completed_queue: The queue in which extracted documentation will
            be placed
    """
    # pylint: disable=import-outside-toplevel
    # pylint: disable=too-many-locals

    # load the fragment_loader _after_ the path is set
    from ansible.plugins.loader import fragment_loader  # noqa: PLC0415

    while True:
        entry = pending_queue.get()
        if entry is None:
            break
        collection_name, checksum, plugin_path = entry

        try:
            if ansible_version.startswith("2.9"):
                (doc, examples, returndocs, metadata) = get_docstring(
                    filename=str(plugin_path),
                    fragment_loader=fragment_loader,
                )
            else:
                (doc, examples, returndocs, metadata) = get_docstring(
                    filename=str(plugin_path),
                    fragment_loader=fragment_loader,
                    collection_name=collection_name,
                )

        except Exception:  # noqa: BLE001
            try:
                with plugin_path.open(mode="r", encoding="utf-8") as f:
                    content = f.read()
                doc, examples, returndocs, metadata = get_doc_withast(content)
                doc, examples, returndocs, metadata = (
                    yaml.load(value, Loader=yaml.SafeLoader)
                    for value in (doc, examples, returndocs, metadata)
                )
            except Exception as exc:  # noqa: BLE001
                err_message = f"{type(exc).__name__} (get_docstring): {exc!s}"
                completed_queue.put(("error", (checksum, plugin_path, err_message)))
                continue

        try:
            q_message = {
                "plugin": {
                    "doc": doc,
                    "examples": examples,
                    "returndocs": returndocs,
                    "metadata": metadata,
                },
                "timestamp": datetime.now(timezone.utc).isoformat(),
            }
            completed_queue.put(("plugin", (checksum, json.dumps(q_message, default=str))))
        except JSONDecodeError as exc:
            err_message = f"{type(exc).__name__} (json_decode_doc): {exc!s}"
            completed_queue.put(("error", (checksum, plugin_path, err_message)))


def identify_missing(
    collections: dict[Any, Any],
    collection_cache: KeyValueStore,
) -> tuple[set[Any], list[Any], int]:
    """Identify plugins missing from the cache.

    Args:
        collections: All plugins found across all collections
        collection_cache: The key value interface to a sqlite database

    Returns:
        Handled and plugins missing from the cache, including a count of
        plugins
    """
    handled = set()
    missing = []
    plugin_count = 0
    for collection in collections.values():
        for checksum, details in collection["plugin_checksums"].items():
            plugin_count += 1
            if checksum not in handled:
                if checksum not in collection_cache:
                    missing.append(
                        (
                            collection["known_as"],
                            checksum,
                            Path(collection["path"], details["path"]),
                        ),
                    )
                handled.add(checksum)
    return handled, missing, plugin_count


def parse_args() -> tuple[argparse.Namespace, list[Path]]:
    # pylint: disable=possibly-used-before-assignment
    """Parse the arguments from the command line.

    Returns:
        The parsed arguments and all directories to search
    """
    parser = argparse.ArgumentParser(description="Catalog collections.")
    parser.add_argument(
        "-d",
        dest="dirs",
        nargs="+",
        help="search within the specified directories",
        default=current_collection_paths,
    )

    parser.add_argument("-a", dest="adjacent", help="prepended to dirs")
    parser.add_argument(
        "-c",
        dest="collection_cache_path",
        help="path to collection cache",
        required=True,
    )
    parsed_args = parser.parse_args()

    adjacent = vars(parsed_args).get("adjacent")
    directories = [adjacent] + parsed_args.dirs if adjacent else parsed_args.dirs

    directories.extend(reversed(sys.path))

    resolved = []
    for directory in directories:
        realpath = Path(directory).resolve()
        if realpath not in resolved:
            resolved.append(realpath)

    return parsed_args, resolved


def retrieve_collections_paths() -> dict[Any, Any]:
    """Retrieve the currently set collection paths.

    Returns:
        Errors or the configured collection directories
    """
    cmd = ["ansible-config", "dump"]
    proc_out = run_command(cmd)
    if "error" in proc_out:
        return proc_out

    # Filter for COLLECTIONS_PATHS instead of using shell pipe and grep
    collections_paths_line = None
    for line in proc_out["stdout"].splitlines():
        if "COLLECTIONS_PATHS" in line:
            collections_paths_line = line
            break

    if not collections_paths_line:
        return {"error": "COLLECTIONS_PATHS not found in ansible-config output"}

    regex = re.compile(r"^(?P<variable>\S+)\((?P<source>.*)\)\s=\s(?P<current>.*)$")
    parsed = regex.match(collections_paths_line)
    if parsed:
        try:
            current = yaml.load(parsed.groupdict()["current"], Loader=SafeLoader)
        except (YAMLError, KeyError) as exc:
            return {"error": str(exc)}
        return {"result": current}
    return {"error": f"corrupt current collection path: {collections_paths_line}"}


def retrieve_docs(
    collection_cache: KeyValueStore,
    errors: list[dict[str, str]],
    missing: list[str],
    stats: dict[Any, Any],
) -> None:
    # pylint: disable=too-many-locals
    """Extract the docs from the plugins.

    Args:
        collection_cache: The key value interface to a sqlite database
        errors: Previous errors encountered
        missing: Plugins missing from the collection cache
        stats: Statistics related to the collection cataloging process
    """
    pending_queue = multiprocessing.Manager().Queue()
    completed_queue = multiprocessing.Manager().Queue()
    processes = []
    for _proc in range(PROCESSES):
        proc = multiprocessing.Process(target=worker, args=(pending_queue, completed_queue))
        processes.append(proc)
        proc.start()

    for entry in missing:
        pending_queue.put(entry)
    for _proc in range(PROCESSES):
        pending_queue.put(None)
    for proc in processes:
        proc.join()

    while not completed_queue.empty():
        message_type, message = completed_queue.get()
        if message_type == "plugin":
            checksum, plugin = message
            collection_cache[checksum] = plugin
            stats["cache_added_success"] += 1
        elif message_type == "error":
            checksum, plugin_path, error = message
            collection_cache[checksum] = json.dumps({"error": error})
            errors.append({"path": str(plugin_path), "error": error})
            stats["cache_added_errors"] += 1


def get_doc_withast(content: Any) -> tuple[Any, Any, Any, Any]:
    """Get the documentation, examples, returndocs, and metadata from the content using ast.

    Args:
        content: The content to parse using ast.

    Returns:
        A tuple containing the documentation, examples, returndocs, and
        metadata.
    """
    doc, examples, returndocs, metadata = "", "", "", ""

    # Parse the content using ast and walk through nodes
    for node in ast.walk(ast.parse(content)):
        if isinstance(node, ast.Assign) and isinstance(node.targets[0], ast.Name):
            target_id = node.targets[0].id

            # Check if node.value is an instance of ast.Constant and its value is a string
            if isinstance(node.value, ast.Constant) and isinstance(node.value.value, str):
                if target_id == "DOCUMENTATION":
                    doc = node.value.value.strip()
                elif target_id == "EXAMPLES":
                    examples = node.value.value.strip()
                elif target_id == "RETURN":
                    returndocs = node.value.value.strip()
                elif target_id == "METADATA":
                    metadata = node.value.value.strip()
    return doc, examples, returndocs, metadata


def run_command(cmd: list[str]) -> dict[str, str]:
    """Run a command using subprocess.

    Args:
        cmd: The command to run, split

    Returns:
        Errors and the stdout from the command run
    """
    try:
        proc_out = subprocess.run(
            cmd,
            capture_output=True,
            check=True,
            text=True,
            shell=False,
        )
    except subprocess.CalledProcessError as exc:
        return {"error": str(exc)}
    except (FileNotFoundError, PermissionError) as exc:
        return {"error": f"{type(exc).__name__}: {exc}"}
    return {"stdout": proc_out.stdout}


def main() -> dict[Any, Any]:
    # pylint: disable=possibly-used-before-assignment
    """Run the collection catalog process.

    Returns:
        The results from the completed collection cataloging process
    """
    stats = {}
    stats["cache_added_success"] = 0
    stats["cache_added_errors"] = 0

    cc_obj = CollectionCatalog(directories=parent_directories)
    collections, errors = cc_obj.process_directories()
    stats["collection_count"] = len(collections)

    collection_cache_path = Path(args.collection_cache_path).resolve().expanduser()
    collection_cache = KeyValueStore(collection_cache_path)

    handled, missing, plugin_count = identify_missing(collections, collection_cache)
    stats["plugin_count"] = plugin_count
    stats["unique plugins"] = len(handled)
    stats["processed"] = len(missing)

    if missing:
        retrieve_docs(collection_cache, errors, missing, stats)

    cached_checksums = collection_cache.keys()
    stats["cache_length"] = len(collection_cache.keys())

    for collection in collections.values():
        for no_doc in set(collection["plugin_checksums"].keys()) - set(cached_checksums):
            del collection["plugin_checksums"][no_doc]

    collection_cache.close()
    return {
        "collections": collections,
        "errors": errors,
        "stats": stats,
        "messages": cc_obj._messages,  # noqa: SLF001
    }


if __name__ == "__main__":
    start_time = datetime.now(timezone.utc)

    collection_paths = retrieve_collections_paths()
    if "error" in collection_paths:
        sys.exit(collection_paths["error"])
    else:
        current_collection_paths = collection_paths["result"]

    args, parent_directories = parse_args()

    COLLECTION_SCAN_PATHS = ":".join(str(parent) for parent in parent_directories)
    os.environ["ANSIBLE_COLLECTIONS_PATH"] = COLLECTION_SCAN_PATHS

    result = main()
    result["stats"]["duration"] = (datetime.now(timezone.utc) - start_time).total_seconds()
    result["collection_scan_paths"] = COLLECTION_SCAN_PATHS
    print(json.dumps(result, default=str))



"""A package containing all available actions.

The ``actions`` package is a plugin-like implementation for all available application actions.

Actions can be added without registration and are discovered and loaded
when their ``KEGEX`` is matched against user input.

This allows for each action to be loaded multiple times, creating a unique
and isolated instance of the action.

Currently, the ``actions`` package is the only package supported for actions and
is identified in the
:class:`~ansible_navigator.configuration_subsystem.navigator_configuration.Internals`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from . import _actions as actions


if TYPE_CHECKING:
    from collections.abc import Callable

    from ansible_navigator.action_defs import RunStdoutReturn
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
    from ansible_navigator.ui_framework import Interaction


get: Callable[[str], Any] = actions.get_factory(__package__)

names = actions.names_factory(__package__)

kegexes: Callable[..., Any] = actions.kegexes_factory(__package__)

run_action_stdout: Callable[
    [str, ApplicationConfiguration],
    RunStdoutReturn,
] = actions.run_stdout_factory(
    __package__,
)

run_action: Callable[
    [str, AppPublic, Interaction],
    Interaction | None,
] = actions.run_interactive_factory(__package__)


__all__ = ["get", "kegexes", "names", "run_action", "run_action_stdout"]



"""The settings file migration base class."""

from __future__ import annotations

from typing import Any

from ansible_navigator.content_defs import ContentView
from ansible_navigator.content_defs import SerializationFormat
from ansible_navigator.utils.ansi import COLOR
from ansible_navigator.utils.ansi import changed
from ansible_navigator.utils.ansi import info
from ansible_navigator.utils.serialize import Loader
from ansible_navigator.utils.serialize import serialize_write_file
from ansible_navigator.utils.serialize import yaml

from .definitions import Migration


class SettingsFile(Migration):
    """The settings file migration base class."""

    name = "Settings file migration base class"

    def __init__(self) -> None:
        """Initialize the settings file migration."""
        super().__init__()
        self.content: dict[Any, Any] = {}
        self._backup_suffix = ".v0"

    def run(self, *args: Any, **kwargs: Any) -> None:
        """Perform the settings file migration.

        Args:
            *args: Positional arguments
            **kwargs: Keyword arguments
        """
        # pylint: disable=possibly-used-before-assignment
        if not self.content:
            with self.settings_file_path.open("r", encoding="utf-8") as f:
                try:
                    self.content = yaml.load(f, Loader=Loader)
                except (yaml.scanner.ScannerError, yaml.parser.ParserError):
                    return

        if not self.content:
            return

        # Check if any of the migrations are needed
        if self.check:
            self.run_steps()
            self.was_needed = self.needed_now
            return

        # Not check and wasn't needed
        if not self.was_needed:
            return

        self.run_steps()

        # Something may have gone wrong
        if self.needed_now:
            return

        # Back up the current
        backup = self.settings_file_path.rename(
            self.settings_file_path.with_suffix(self._backup_suffix),
        )
        info(color=COLOR, message=f"Backup: {backup}")

        # Write the new file
        if self.settings_file_path.suffix in (".yml", ".yaml"):
            serialization_format = SerializationFormat.YAML
        elif self.settings_file_path.suffix == ".json":
            serialization_format = SerializationFormat.JSON

        serialize_write_file(
            content=self.content,
            content_view=ContentView.NORMAL,
            file_mode="w",
            file=self.settings_file_path,
            serialization_format=serialization_format,
        )
        changed(color=COLOR, message=f"Updated: {self.settings_file_path}")



from __future__ import annotations

from typing import TYPE_CHECKING

from .region import Region
from .region import Regions
from .state import State


if TYPE_CHECKING:
    from .compiler import Compiler


def tokenize(
    compiler: Compiler,
    state: State,
    line: str,
    first_line: bool,
) -> tuple[State, Regions]:
    """Tokenize a string into it's parts."""
    ret: list[Region] = []
    pos = 0
    boundary = state.cur.boundary

    # this is still a little wasteful
    while_stack = []
    for while_rule, idx in state.while_stack:
        while_stack.append((while_rule, idx))
        while_state = State(state.entries[:idx], tuple(while_stack))

        while_res = while_rule.continues(compiler, while_state, line, pos, first_line, boundary)
        if while_res is None:
            state = while_state.pop_while()
            break
        pos, boundary, regions = while_res
        ret.extend(regions)

    search_res = state.cur.rule.search(compiler, state, line, pos, first_line, boundary)
    while search_res is not None:
        state, pos, boundary, regions = search_res
        ret.extend(regions)

        search_res = state.cur.rule.search(compiler, state, line, pos, first_line, boundary)

    if pos < len(line):
        ret.append(Region(pos, len(line), state.cur.scope))

    return state, tuple(ret)



from typing import Generic
from typing import TypeVar

from ._types import Protocol


TKey_contra = TypeVar("TKey_contra", contravariant=True)
TValue_co = TypeVar("TValue_co", covariant=True)


class Indexable(Generic[TKey_contra, TValue_co], Protocol):  # noqa: PYI059
    def __getitem__(self, key: TKey_contra) -> TValue_co:
        """Get the value associated with the given key.

        Args:
            key: The key to retrieve the value for.
        """


class FChainMap(Generic[TKey_contra, TValue_co]):
    def __init__(self, *mappings: Indexable[TKey_contra, TValue_co]) -> None:
        """Initialize the FChainMap.

        Args:
            *mappings: A tuple of mappings, each corresponding to a
                repository entry of a ``tm_language`` file
        """
        self._mappings = mappings

    def __getitem__(self, key: TKey_contra) -> TValue_co:
        """Get the value associated with the given key.

        Args:
            key: The key to retrieve the value for.

        Returns:
            The value associated with the key.

        Raises:
            KeyError: If the key is not found in any of the mappings.
        """
        for mapping in reversed(self._mappings):
            try:
                return mapping[key]
            except KeyError:
                pass
        raise KeyError(key)



"""``:rerun`` command implementation."""

import copy
import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:rerun`` command implementation."""

    KEGEX = r"^rr|rerun?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:rerun`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:rerun`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("rerun requested")
        this = copy.copy(app.steps.current)
        app.rerun()
        # ensure we are last on the stack
        if app.steps.current != this:
            app.steps.append(this)



"""Public entry points for the runner api."""

from .ansible_config import AnsibleConfig
from .ansible_doc import AnsibleDoc
from .ansible_inventory import AnsibleInventory
from .command import Command
from .command_async import CommandAsync


__all__ = (
    "AnsibleConfig",
    "AnsibleDoc",
    "AnsibleInventory",
    "Command",
    "CommandAsync",
)



"""Implementation of the asynchronous invocation of ``ansible-runner``.

Herein lies the ability to invoke ``ansible-runner`` in an async fashion.
A queue is provided and ``ansible-runner`` uses ``pexpect`` to parse
standard out and error from the command run and populates the
queue with messages.
"""

from copy import deepcopy
from queue import Queue
from threading import Thread
from typing import Any

from ansible_runner import run_command_async

from .command_base import CommandBase


class CommandAsync(CommandBase):
    """A wrapper for the asynchronous runner."""

    def __init__(
        self,
        executable_cmd: str,
        queue: Queue[Any],
        write_job_events: bool,
        **kwargs: Any,
    ) -> None:
        """Initialize the arguments for the ``run_command_async`` interface of ``ansible-runner``.

        For common arguments refer to the documentation of the ``CommandBase`` class.

        Args:
            executable_cmd: The command to be invoked
            queue: The queue to post events from ``ansible-runner``
            write_job_events: Allows job_events to be processed by
                ``ansible-runner``
            **kwargs: The arguments for the async runner call
        """
        self._queue = queue
        self._write_job_events = write_job_events
        super().__init__(executable_cmd, **kwargs)

    def _event_handler(self, event: Any) -> bool:
        """Handle the event from ansible-runner.

        Args:
            event: The event from ansible-runner

        Returns:
            The value of ``self._write_job_events``, a boolean
        """
        self._logger.debug("ansible-runner event handle: %s", event)
        new_event = deepcopy(event)
        self._queue.put(new_event)
        return self._write_job_events

    def run(self) -> Thread:
        """Initiate the execution of the runner command in async mode.

        Returns:
            The runner thread
        """
        self.generate_run_command_args()
        self._runner_args.update({"event_handler": self._event_handler})
        thread, self.ansible_runner_instance = run_command_async(**self._runner_args)
        self.status = "running"
        return thread



# cspell:ignore A_INVIS
"""Tokenize and color text."""

from __future__ import annotations

import colorsys
import copy
import curses
import functools
import json
import logging
import re

from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.tm_tokenize.grammars import Grammars
from ansible_navigator.tm_tokenize.tokenize import tokenize

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_defs import CursesLines
from .curses_defs import RgbTuple
from .curses_defs import SimpleLinePart
from .ui_constants import Color
from .ui_constants import Decoration


if TYPE_CHECKING:
    from ansible_navigator.tm_tokenize.region import Regions
    from ansible_navigator.utils.compatibility import Traversable


CURSES_STYLES = {
    0: None,
    1: getattr(curses, "A_BOLD", None),
    2: getattr(curses, "A_DIM", None),
    3: getattr(curses, "A_ITALIC", None),
    4: getattr(curses, "A_UNDERLINE", None),
    5: getattr(curses, "A_BLINK", None),
    6: getattr(curses, "A_BLINK", None),
    7: getattr(curses, "A_REVERSE", None),
    8: getattr(curses, "A_INVIS", None),
}


class ColorSchema:
    """A storage mechanism for the schema (theme)."""

    def __init__(self, schema: dict[str, str | list[Any] | dict[Any, Any]]) -> None:
        """Initialize the ColorSchema class.

        Args:
            schema: The color scheme, theme to use
        """
        self._logger = logging.getLogger(__name__)
        self._schema = schema

    @functools.lru_cache(maxsize=128)  # noqa: B019
    def get_color_and_style(self, scope: str) -> tuple[RgbTuple | None, str | None]:
        """Get a color from the schema, traverse all to aggregate color and style.

        Args:
            scope: The scope, aka format

        Returns:
            The color in RGB format or nothing
        """
        found_color = None
        found_style = None
        for name in scope:
            for parts in range(len(name.split("."))):
                prop = name.split()[-1].rsplit(".", parts)[0]
                color = next(
                    (
                        token_color
                        for token_color in self._schema["tokenColors"]
                        if (
                            isinstance(token_color, dict)
                            and prop in scope_to_list(token_color.get("scope", []))
                        )
                    ),
                    None,
                )
                if color:
                    found_color = color.get("settings", {}).get("foreground", None)
                    found_style = color.get("settings", {}).get("fontStyle", None)
        if found_color:
            found_color = hex_to_rgb(found_color)
        return found_color, found_style


class Colorize:
    """Functionality for coloring."""

    def __init__(self, grammar_dir: Traversable, theme_path: Traversable) -> None:
        """Initialize the colorizer.

        Args:
            grammar_dir: The directory in which the grammars reside
            theme_path: The path to the currently configured color theme
        """
        self._logger = logging.getLogger(__name__)
        self._schema: ColorSchema
        self._grammars = Grammars(str(grammar_dir))
        self._theme_path = Path(str(theme_path))
        self._load()

    def _load(self) -> None:
        """Load the color scheme from the file system."""
        with self._theme_path.open(mode="r", encoding="utf-8") as fh:
            self._schema = ColorSchema(json.load(fh))

    @staticmethod
    @functools.lru_cache(maxsize=100)
    def render_ansi(doc: str) -> CursesLines:
        """Convert ansi colored text into curses lines.

        Args:
            doc: The text to convert

        Returns:
            Lines ready to present using the TUI
        """
        lines = tuple(ansi_to_curses(line) for line in doc.splitlines())
        return CursesLines(lines)

    @functools.lru_cache(maxsize=100)  # noqa: B019
    def render(self, doc: str, scope: str) -> list[list[SimpleLinePart]]:
        """Render text lines into lines of columns and colors.

        Args:
            doc: The string to split, tokenize and color
            scope: The scope, aka the format of the string

        Returns:
            A list of lines, each a list of dicts
        """
        try:
            compiler = self._grammars.compiler_for_scope(scope)
        except KeyError:
            compiler = None

        if compiler and scope != "no_color":
            state = compiler.root_state
            lines = []
            for line_idx, line in enumerate(doc.splitlines()):
                line += "\n"
                first_line = line_idx == 0
                try:
                    state, regions = tokenize(compiler, state, line, first_line)
                except Exception as exc:  # noqa: BLE001
                    self._logger.critical(
                        (
                            "An unexpected error occurred within the tokenization"
                            " subsystem.  Please log an issue with the following:"
                        ),
                    )
                    self._logger.critical(
                        "  Err: '%s', Scope: '%s', Line follows....",
                        str(exc),
                        scope,
                    )
                    self._logger.critical("  '%s'", line)
                    self._logger.critical("  The current content will be rendered without color")
                    break
                else:
                    lines.append((regions, line))
            else:
                assembled = columns_and_colors(lines, self._schema)
                if scope == "text.html.markdown":
                    assembled = strip_markdown(assembled)
                return assembled

        res = [
            [SimpleLinePart(column=0, chars=doc_line, color=None, style=None)]
            for doc_line in doc.splitlines()
        ]
        return res


def scope_to_list(scope: str | list[Any]) -> list[Any]:
    """Convert a token scope to a list if necessary.

    A scope in a theme should always be a string or list,
    but just in case return an empty list if not

    Args:
        scope: The scope

    Returns:
        Scope as list
    """
    if isinstance(scope, list):
        return scope
    if isinstance(scope, str):
        return [scope]
    return []


def hex_to_rgb(value: str) -> RgbTuple:
    """Convert a hex value to RGB tuple.

    Args:
        value: The hex color

    Returns:
        RGB tuple
    """
    value = value.lstrip("#")
    value_length = len(value)
    red, green, blue = (
        int(value[i : i + value_length // 3], 16) for i in range(0, value_length, value_length // 3)
    )
    return red, green, blue


def scale_for_curses(rgb_value: int) -> int:
    """Scale a single RGB value for curses.

    Args:
        rgb_value: One RGB value

    Returns:
        The value scaled for curses
    """
    curses_ceiling = 1000
    rgb_ceiling = 255
    return int(rgb_value * curses_ceiling / rgb_ceiling)


def hex_to_rgb_curses(value: str) -> RgbTuple:
    """Convert a hex color to RGB scaled for curses.

    Args:
        value: an RGB color

    Returns:
        The colors scaled to 1000
    """
    red, green, blue = hex_to_rgb(value)
    return (scale_for_curses(red), scale_for_curses(green), scale_for_curses(blue))


def _rgb_to_ansi_256(red: int, green: int, blue: int) -> int:
    """Convert an RGB color to a 256-color ansi value.

    Args:
        red: The red component
        green: The green component
        blue: The blue component

    Returns:
        A 256-color ansi value
    """
    if red == green and green == blue:
        if red < 8:
            return 16
        if red > 248:
            return 231
        return round(((red - 8) / 247) * 24) + 232
    return 16 + (36 * round(red / 255 * 5)) + (6 * round(green / 255 * 5)) + round(blue / 255 * 5)


def _rgb_to_ansi_16(red: int, green: int, blue: int) -> int:
    """Convert an RGB color to a 16-color ansi value.

    Args:
        red: The red component
        green: The green component
        blue: The blue component

    Returns:
        A 16-color ansi value
    """
    value = colorsys.rgb_to_hsv(red, green, blue)[2]
    value = round(value / 50)
    if value == 0:
        return 30
    ansi = (round(blue / 255) << 2) | (round(green / 255) << 1) | round(red / 255)
    if value == 2:
        ansi += 8
    return ansi


def rgb_to_ansi(red: int, green: int, blue: int, colors: int) -> int:
    """Convert an RGB color to an ansi color.

    Args:
        red: The red component
        green: The green component
        blue: The blue component
        colors: The number of color supported by the terminal

    Returns:
        A color suitable for the terminal
    """
    # https://github.com/Qix-/color-convert/blob/master/conversions.js
    if colors == 256:
        ansi = _rgb_to_ansi_256(red, green, blue)
    elif colors == 16:
        ansi = _rgb_to_ansi_16(red, green, blue)
    else:  # colors == 8, sorry
        ansi = (round(blue / 255) << 2) | (round(green / 255) << 1) | round(red / 255)
    return ansi


def _apply_regions_to_parts(
    line_parts: list[SimpleLinePart],
    regions: Regions,
    schema: ColorSchema,
) -> None:
    """Apply color and style from regions to line parts.

    Args:
        line_parts: The individual character parts of a line
        regions: The tokenized regions for the line
        schema: An instance of the ColorSchema
    """
    for region in regions:
        color, style = schema.get_color_and_style(region.scope)
        if color:
            for idx in range(region.start, region.end):
                line_parts[idx].color = color
        if style:
            for idx in range(region.start, region.end):
                line_parts[idx].style = style


def _compress_line_parts(
    line_parts: list[SimpleLinePart],
    line_text: str,
) -> list[SimpleLinePart]:
    """Compress adjacent line parts with the same color and style.

    Args:
        line_parts: The individual character parts of a line
        line_text: The original line text

    Returns:
        Compressed line parts
    """
    if not line_parts:
        return [SimpleLinePart(chars=line_text, color=None, column=0, style=None)]
    grouped = [line_parts.pop(0)]
    while line_parts:
        entry = line_parts.pop(0)
        if entry.color == grouped[-1].color and entry.style == grouped[-1].style:
            grouped[-1].chars += entry.chars
        else:
            grouped.append(entry)
    return grouped


def _update_line_part_columns(results: list[list[SimpleLinePart]]) -> None:
    """Update column offsets in each line part based on preceding text lengths.

    Args:
        results: Lines of text parts to update
    """
    for result in results:
        column = 0
        for line_part in result:
            line_part.column = column
            column += len(line_part.chars)


def columns_and_colors(
    lines: list[tuple[Regions, str]],
    schema: ColorSchema,
) -> list[list[SimpleLinePart]]:
    """Convert to colors and columns.

    Args:
        lines: Lines of text and their regions
        schema: An instance of the ColorSchema

    Returns:
        Lines of text, each broken into sections
    """
    results: list[list[SimpleLinePart]] = []

    for line in lines:
        # Break the into 1 character parts, temporarily set the column to 0
        line_parts = [
            SimpleLinePart(chars=character, color=None, column=0, style=None)
            for character in line[1]
        ]

        # Replace the color with the RgbTuple
        _apply_regions_to_parts(line_parts, line[0], schema)

        # Compress the line, grouping characters of like color and decoration
        results.append(_compress_line_parts(line_parts, line[1]))

    # Update the column in each line part, based on the total of the preceding text lengths
    _update_line_part_columns(results)

    return results


def _parse_ansi_match(
    cap: dict[str, str | None],
    color: int = 0,
    style: int = 0,
) -> tuple[int, int]:
    """Parse an ANSI color match groupdict into a color and style.

    Args:
        cap: The groupdict from a color_regex match
        color: The current color state to preserve across calls
        style: The current style state to preserve across calls

    Returns:
        A (color, style) tuple
    """
    one = cap["one"]
    two = cap["two"]
    if cap["fg_action"] == "39;" or (one == "0" and two is None):
        pass  # default color
    elif cap["fg_action"] == "38;5;":
        color = int(one)  # type: ignore[arg-type]
        if two:
            style = CURSES_STYLES.get(int(two)) or 0
    elif not cap["fg_action"]:
        ansi_16 = list(chain(range(30, 38), range(90, 98)))
        if two is None:
            color = ansi_16.index(int(one)) if int(one) in ansi_16 else int(one)  # type: ignore[arg-type]
        else:
            color = ansi_16.index(int(two)) if int(two) in ansi_16 else int(two)
            style = CURSES_STYLES.get(int(one)) or 0  # type: ignore[arg-type]
    return color, style


def ansi_to_curses(line: str) -> CursesLine:
    """Convert ansible color codes to curses colors.

    Args:
        line: A string with ansi colors

    Returns:
        A line ready for presentation in the TUI
    """
    if line == "":
        line_part = CursesLinePart(
            column=0,
            string="",
            color=Color.BLACK,
            decoration=Decoration.NORMAL,
        )
        return CursesLine((line_part,))

    printable = []
    ansi_regex = re.compile(r"(\x1b\[[\d;]*m)")
    color_regex = re.compile(
        r"""(?x)
            \x1b\[                              # Control Sequence Introducer
            (?P<fg_action>(38;5|39);)?          # optional FG color action
            (?P<_bg_action>(48;5|49);)?         # optional BG color action
            (?P<one>\d+)                        # required, one number
            (;(?P<two>\d+))?                    # optional 2nd number
            m
        """,
    )
    parts = ansi_regex.split(line)
    colno = 0
    color = 0
    style = 0
    while parts:
        part = parts.pop(0)
        if part:
            match = color_regex.match(part)
            if match:
                color, style = _parse_ansi_match(match.groupdict(), color, style)
            else:
                curses_line = CursesLinePart(
                    column=colno,
                    string=part,
                    color=color,
                    decoration=style,
                )
                printable.append(curses_line)
                colno += len(part)
                color = 0
                style = 0
    return CursesLine(tuple(printable))


def _process_markdown_part(
    lines: list[list[SimpleLinePart]],
    line_idx: int,
    part_idx: int,
    part: SimpleLinePart,
    in_a_code_block: bool,
    full_dash_line: list[SimpleLinePart],
) -> bool:
    """Process a single markdown part and apply transformations.

    Args:
        lines: Lines of text and their parts (mutated in place)
        line_idx: The index of the current line
        part_idx: The index of the current part within the line
        part: The current part to process
        in_a_code_block: Whether we are currently inside a code block
        full_dash_line: The dash line to use for separators

    Returns:
        The updated in_a_code_block state
    """
    if part.chars.startswith("```"):
        # Remove ```x from a line
        lines[line_idx][part_idx].chars = "\n"
        return not in_a_code_block

    if in_a_code_block:
        # Don't modify inside a code block
        return in_a_code_block

    if part.chars.startswith("#"):
        # Remove # headings
        lines[line_idx][part_idx].chars = re.sub(r"^(#{1,6}\s)(.*$)", r"\2", part.chars)
        if part.chars.startswith("# "):
            # Insert a full line after a heading 1
            lines.insert(line_idx + 1, full_dash_line)
        return in_a_code_block

    if part.chars == "---\n":
        # Replace all dash line with no-break dashes if \n before and after
        try:
            if lines[line_idx - 1][0].chars == "\n" and lines[line_idx + 1][0].chars == "\n":
                lines[line_idx] = full_dash_line
                return in_a_code_block
        except IndexError:
            pass

    if "`" in part.chars:
        # Remove `` from code blocks
        lines[line_idx][part_idx].chars = re.sub(r"`(.*)`", r"\1", part.chars)

    if "*" in part.chars:
        # Remove `` from code blocks
        lines[line_idx][part_idx].chars = re.sub(r"\*(.*)\*", r"\1", part.chars)

    return in_a_code_block


def strip_markdown(lines: list[list[SimpleLinePart]]) -> list[list[SimpleLinePart]]:
    """Strip some markdown from the regions.

    This is not a complete removal of markdown, but it removes some of the
    common markdown that is in use.

    Args:
        lines: Lines of text and their parts

    Returns:
        Lines of text and their parts without some markdown
    """
    full_dash_line = [
        SimpleLinePart(
            chars=f"{'—' * 132}\n",
            column=0,
            color=(128, 128, 128),
            style=None,
        ),
    ]
    in_a_code_block = False
    for line_idx, line in reversed(list(enumerate(copy.deepcopy(lines)))):
        for part_idx, part in enumerate(line):
            in_a_code_block = _process_markdown_part(
                lines,
                line_idx,
                part_idx,
                part,
                in_a_code_block,
                full_dash_line,
            )

    return lines



"""ANSI color codes and functions for terminal output."""

import os
import sys

from sys import stdout

from .definitions import Color


IS_TTY = stdout.isatty()
COLOR = "NO_COLOR" not in os.environ and IS_TTY


def changed(color: bool, message: str) -> None:
    """Output changed information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"\r{Color.YELLOW}{message}{Color.END}\033[K")
    else:
        print(message)


def failed(color: bool, message: str) -> None:
    """Output failure information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"\r{Color.RED}{message}{Color.END}\033[K")
    else:
        print(message)


def info(color: bool, message: str) -> None:
    """Output info information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"{Color.CYAN}{message}{Color.END}")
    else:
        print(message)


def subtle(color: bool, message: str) -> None:
    """Output subtle information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"{Color.GREY}{message}{Color.END}")
    else:
        print(message)


def prompt_enter() -> None:
    """Output prompt information to the console."""
    try:
        input("Press Enter to continue: ")
    except KeyboardInterrupt:
        sys.exit(0)


def prompt_yn(message: str) -> bool:
    """Output prompt information to the console.

    Args:
        message: The message to output

    Returns:
        Whether the user answered yes
    """
    try:
        reply = None
        while reply not in ("", "y", "n"):
            reply = input(f"{message} (Y/n): ").lower()
    except KeyboardInterrupt:
        sys.exit(0)
    else:
        return reply in ("", "y")


def success(color: bool, message: str) -> None:
    """Output success information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"\r{Color.GREEN}{message}{Color.END}\033[K")
    else:
        print(message)


def warning(color: bool, message: str) -> None:
    """Output warning information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"{Color.YELLOW}{message}{Color.END}")
    else:
        print(message)


def working(color: bool, message: str) -> None:
    """Output working information to the console.

    Args:
        color: Whether to color the message
        message: The message to output
    """
    if color:
        print(f"{Color.GREY}{message}{Color.END}", end="", flush=True)
    else:
        print(message)


def blank_line() -> None:
    """Output a blank line to the console."""
    print()



"""``:sample_form`` command implementation."""

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import dict_to_form
from ansible_navigator.ui_framework import form_to_dict
from ansible_navigator.utils.serialize import yaml

from . import _actions as actions


FORM = """
form:
  title:  Please confirm the following information
  fields:
  - name: construct
    prompt: Please name an Ansible construct
    type: text_input
    validator:
      name: one_of
      choices:
      - collection
      - playbook
      - role
      - plugin
  - name: ansible_fest
    prompt: In 2019, AnsibleFest was held in what city?
    default: Atlanta
    type: text_input
    validator:
      name: none
  - name: what_is
    prompt: In Rocannon's World, an "ansible" is an instantaneous communication device
    type: text_input
    validator:
      name: yes_no
  - name: rh_acquire
    prompt: What year did Red Hat acquire Ansible
    default: 2015
    type: text_input
    validator:
      name: something
  - name: ansible_24
    prompt: Was Ansible 2.4 released in 2018
    type: text_input
    validator:
      name: true_false
  - name: fc_result
    type: checkbox
    prompt: Ansible can automate
    options:
    - name: clouds
      text: Clouds
    - name: lightbulbs
      text: Lightbulbs
    - name: linux
      text: Linux servers
    - name: network
      text: Networks
      checked: True
    - name: windows
      text: Windows servers
    - name: nothing
      text: Nothing
      disabled: True
    max_selected: 4
  - name: fr_result
    type: radio
    prompt: The most popular network module is
    options:
    - name: cli_command
      text: ansible.netcommon.cli_command
    - name: cli_config
      text: ansible.netcommon.cli_config
    - name: nxos_interfaces
      text: cisco.nxos.nxos_interfaces
  - name: file_name
    type: text_input
    prompt: Provide a valid file path
    validator:
      name: valid_file_path
    pre_populate: /etc/hostname
"""


@actions.register
class Action(ActionBase):
    """``:sample_form`` command implementation."""

    KEGEX = r"^sample_form$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:sample_form`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="sample_form")

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction:
        """Execute the ``:sample_form`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction`
        """
        self._logger.debug("sample form requested")
        self._prepare_to_run(app, interaction)

        form_data = yaml.safe_load(FORM)
        form = dict_to_form(form_data["form"])
        interaction.ui.show_form(form)
        as_dict = form_to_dict(form)
        self._logger.debug("form response: %s", as_dict)

        while True:
            self._calling_app.update()
            next_interaction: Interaction = interaction.ui.show(obj=as_dict)
            if next_interaction.name != "refresh":
                break

        self._prepare_to_exit(interaction)
        return next_interaction
